"""Package related to goth's probe interface."""

import abc
import asyncio
from collections import OrderedDict
import contextlib
import copy
import logging
from pathlib import Path
from typing import (
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

from docker import DockerClient

from goth.address import (
    YAGNA_BUS_URL,
    YAGNA_REST_URL,
)

from goth import gftp
from goth.node import DEFAULT_SUBNET
from goth.runner import process
from goth.runner.cli import Cli, YagnaDockerCli
from goth.runner.container.utils import get_container_address
from goth.runner.container.yagna import (
    YagnaContainer,
    YagnaContainerConfig,
    PAYMENT_MOUNT_PATH,
)
from goth.runner.exceptions import KeyAlreadyExistsError
from goth.runner.log import LogConfig, monitored_logger
from goth.runner.log_monitor import PatternMatchingEventMonitor
from goth.runner.probe.agent import AgentComponent, ProviderAgentComponent
from goth.runner.probe.mixin import (
    ActivityApiMixin,
    MarketApiMixin,
    PaymentApiMixin,
    ProviderLogMixin,
)
from goth.runner.probe.rest_client import RestApiComponent


if TYPE_CHECKING:
    from goth.runner import Runner

logger = logging.getLogger(__name__)


class ProbeLoggingAdapter(logging.LoggerAdapter):
    """Adds probe name information to log messages."""

    EXTRA_PROBE_NAME = "probe_name"

    def process(self, msg, kwargs):
        """Process the log message."""
        return "[%s] %s" % (self.extra[self.EXTRA_PROBE_NAME], msg), kwargs


class Probe(abc.ABC):
    """Provides a unified interface for interacting with and testing a single Yagna node.

    This interface consists of several independent modules which may be extended
    in subclasses (see `ProviderProbe` and `RequestorProbe`).
    """

    api: RestApiComponent
    """Component with clients for all three yagna REST APIs."""

    runner: "Runner"
    """A runner that created this probe."""

    cli: YagnaDockerCli
    """A module which enables calling the Yagna CLI on the daemon being tested."""

    container: YagnaContainer
    """A module which handles the lifecycle of the daemon's Docker container."""

    ip_address: Optional[str] = None
    """An IP address of the daemon's container in the Docker network."""

    _agents: "OrderedDict[str, AgentComponent]"
    """Collection of agent components that will be started as part of this probe.

    Keys are agent names, values are subclasses of `AgentComponent`.
    """

    _docker_client: DockerClient
    """A docker client used to create the deamon's container."""

    _gftp_script_dir: Path
    """Directory containing the `gftp` proxy script.

    This script forwards JSON RPC requests to the `gftp` binary running in the docker
    container managed by this probe, and returns responses from the binary.
    """

    _yagna_config: YagnaContainerConfig
    """Config object used for setting up the Yagna node for this probe."""

    def __init__(
        self,
        runner: "Runner",
        client: DockerClient,
        config: YagnaContainerConfig,
        log_config: LogConfig,
    ):
        self.runner = runner
        self._agents = OrderedDict()
        self._docker_client = client
        self._logger = ProbeLoggingAdapter(
            logger, {ProbeLoggingAdapter.EXTRA_PROBE_NAME: config.name}
        )
        config = self._setup_gftp_proxy(config)
        self.container = YagnaContainer(client, config, log_config)
        self.cli = Cli(self.container).yagna
        self._yagna_config = config

    def __str__(self):
        return self.name

    @property
    def agents(self) -> List[AgentComponent]:
        """List of agent components that will be started as part of this probe."""
        return list(self._agents.values())

    @property
    def address(self) -> Optional[str]:
        """Return address from id marked as default."""
        identity = self.cli.id_show()
        return identity.address if identity else None

    @property
    def app_key(self) -> Optional[str]:
        """Return first app key on the list."""
        keys = self.cli.app_key_list()
        return keys[0].key if keys else None

    @property
    def name(self) -> str:
        """Name of the container."""
        return self.container.name

    def _setup_gftp_proxy(self, config: YagnaContainerConfig) -> YagnaContainerConfig:
        """Create a proxy script and a dir for exchanging files with the container."""

        self._gftp_script_dir, gftp_volume_dir = gftp.create_gftp_dirs(config.name)
        self._logger.info("Created gftp script at %s", self._gftp_script_dir)

        new_config = copy.deepcopy(config)
        new_config.volumes[gftp_volume_dir] = gftp.CONTAINER_MOUNT_POINT
        self._logger.info(
            "Gftp volume %s will be mounted at %s in the container",
            gftp_volume_dir,
            gftp.CONTAINER_MOUNT_POINT,
        )

        return new_config

    def add_agent(self, agent: AgentComponent) -> None:
        """Add an agent to be started for this probe."""
        if self._agents.get(agent.name):
            raise KeyAlreadyExistsError(
                f"Probe already has agent component with name: `{agent.name}`"
            )
        self._agents[agent.name] = agent

    async def start(self) -> None:
        """Start the probe."""

        await self._start_container()
        self.api = RestApiComponent(self)

    async def start_agents(self):
        """Start all of the probe's agents."""
        for agent in self.agents:
            await agent.start()

    async def stop(self):
        """
        Stop the probe, removing the Docker container of the daemon being tested.

        Once stopped, a probe cannot be restarted.
        """
        self._logger.info("Stopping probe")
        for agent in self.agents:
            await agent.stop()
        if self.container.logs:
            await self.container.logs.stop()

    def remove(self) -> None:
        """Remove the underlying container."""
        if self.container:
            self.container.remove(force=True)
            self._logger.debug("Container removed")

    async def _start_container(self) -> None:
        """
        Start the probe's Docker container.

        Performs all necessary steps to make the daemon ready for testing
        (e.g. creating the default app key).
        """
        self.container.start()

        # Wait until the daemon is ready to create an app key.
        self._logger.info("Waiting for connection to ya-sb-router")
        if self.container.logs:
            await self.container.logs.wait_for_entry(
                ".*connected with server: ya-sb-router.*", timeout=30
            )
        await self.create_app_key()

        self._logger.info("Waiting for yagna REST API to be listening")
        if self.container.logs:
            await self.container.logs.wait_for_entry(
                "Starting .* service on .*.", timeout=30
            )

        # Obtain the IP address of the container
        self.ip_address = get_container_address(
            self._docker_client, self.container.name
        )
        self._logger.info("IP address: %s", self.ip_address)

    async def create_app_key(self, key_name: str = "test_key") -> str:
        """Attempt to create a new app key on the Yagna daemon.

        The key name can be specified via `key_name` parameter.
        Return the key as string.

        When `self.key_file` is set, this method also:
        - creates ID based on `self.key_file`
        - sets this new ID as default
        - restarts the container ( https://github.com/golemfactory/yagna/issues/458 )
        """
        address = None
        if self._yagna_config.payment_id:
            key_name = self._yagna_config.payment_id.key_file.name
            key_file: str = str(PAYMENT_MOUNT_PATH / key_name)
            self._logger.debug("create_id(alias=%s, key_file=%s", key_name, key_file)
            try:
                db_id = self.cli.id_create(alias=key_name, key_file=key_file)
                address = db_id.address
                self._logger.debug("create_id. alias=%s, address=%s", db_id, address)
            except KeyAlreadyExistsError as e:
                logger.critical("Id already exists : (%r)", e)
                raise
            db_id = self.cli.id_update(address, set_default=True)
            self._logger.debug("update_id. result=%r", db_id)
            self.container.restart()
            await asyncio.sleep(5)
        try:
            key = self.cli.app_key_create(key_name)
            self._logger.debug("create_app_key. key_name=%s, key=%s", key_name, key)
        except KeyAlreadyExistsError:
            app_key = next(
                filter(lambda k: k.name == key_name, self.cli.app_key_list())
            )
            key = app_key.key
        return key

    def set_agent_env_vars(self, env: Dict[str, str]) -> None:
        """Add vars needed to talk to the daemon in this probe's container to `env`."""

        if not self.app_key:
            raise AttributeError("Yagna application key is not set yet")
        path_var = env.get("PATH")
        env.update(
            {
                "YAGNA_APPKEY": self.app_key,
                "YAGNA_API_URL": YAGNA_REST_URL.substitute(host=self.ip_address),
                "GSB_URL": YAGNA_BUS_URL.substitute(host=self.ip_address),
                "PATH": f"{self._gftp_script_dir}:{path_var}",
            }
        )

    @contextlib.asynccontextmanager
    async def run_command_on_host(
        self,
        command: str,
        env: Optional[Dict[str, str]] = None,
        command_timeout: float = 300,
    ) -> Iterator[Tuple[asyncio.Task, PatternMatchingEventMonitor]]:
        """Run `command` on host in given `env` and with optional `timeout`.

        The command is run in the environment extending `env` with variables needed
        to communicate with the daemon running in this probe's container.

        Internally, this method uses `process.run_command()` to run `command`.
        The argument `command_timeout` is passed as the `timeout` parameter to
        `process.run_command()`.

        Returns the `asyncio` task that logs output from the command, and an event
        monitor that observes lines of output produced by the command.

        The task can be awaited in order to wait until the command completes.
        The monitor can be used for asserting properties of the command's output.
        """
        cmd_env = {**env} if env is not None else {}
        self.set_agent_env_vars(cmd_env)

        cmd_monitor = PatternMatchingEventMonitor(name="command output")
        cmd_monitor.start()

        try:
            with monitored_logger(
                f"goth.{self.name}.command_output", cmd_monitor
            ) as cmd_logger:
                cmd_task = asyncio.create_task(
                    process.run_command(
                        command.split(),
                        cmd_env,
                        log_level=logging.INFO,
                        cmd_logger=cmd_logger,
                        timeout=command_timeout,
                    )
                )
                yield cmd_task, cmd_monitor

        except Exception as e:
            logger.warning(f"Cancelling command on error: {e!r}")
            if cmd_task and not cmd_task.done():
                cmd_task.cancel()
            raise

        finally:
            await cmd_monitor.stop()
            logger.debug("Waiting for the command to finish")
            await asyncio.gather(cmd_task, return_exceptions=True)


@contextlib.contextmanager
def create_probe(
    runner: "Runner",
    docker_client: DockerClient,
    config: YagnaContainerConfig,
    log_config: LogConfig,
) -> Iterator[Probe]:
    """Implement a ContextManager protocol for creating and removing probes."""

    probe: Optional[Probe] = None
    try:
        logger.debug(
            "Creating probe. config=%s, probe_type=%s", config, config.probe_type
        )
        probe = config.probe_type(runner, docker_client, config, log_config)
        for name, value in config.probe_properties.items():
            probe.__setattr__(name, value)
        yield probe
    finally:
        if probe:
            logger.debug("Removing probe. name=%s", probe.name)
            probe.remove()


@contextlib.asynccontextmanager
async def run_probe(probe: Probe) -> AsyncIterator[str]:
    """Implement AsyncContextManager for starting and stopping a probe."""

    try:
        logger.debug("Starting probe. name=%s", probe.name)
        await probe.start()
        assert probe.ip_address
        yield probe.ip_address
    finally:
        await probe.stop()


class RequestorProbe(ActivityApiMixin, MarketApiMixin, PaymentApiMixin, Probe):
    """A probe subclass with activity API steps and requestor payment init."""

    async def _start_container(self) -> None:
        await super()._start_container()

        self.cli.payment_fund()
        self.cli.payment_init(sender_mode=True)


class ProviderProbe(MarketApiMixin, PaymentApiMixin, ProviderLogMixin, Probe):
    """A probe subclass that can run a provider agent."""

    provider_agent: ProviderAgentComponent
    """The agent component running `ya-provider` for this probe.
    This field is added for convenience to make getting this agent instance easier."""

    def __init__(
        self,
        runner: "Runner",
        client: DockerClient,
        config: YagnaContainerConfig,
        log_config: LogConfig,
        agent_preset: Optional[str] = None,
        subnet: str = DEFAULT_SUBNET,
    ):
        super().__init__(runner, client, config, log_config)
        self.provider_agent = ProviderAgentComponent(self, subnet, agent_preset)
        self.add_agent(self.provider_agent)



"""A module that provides a wrapper for `gftp` binary running in a docker container.

Note: `DEBUG_GFTP` is an environment variable used by `yapapi` to enable debug logs
from its gftp module. We use the same variable here to set logging level to `DEBUG`
in this wrapper script.

For testing you can use an example request:

  {"jsonrpc": "2.0", "method": "version", "params": {}, "id":1}

The response should be something like:

  {"jsonrpc":"2.0","id":1,"result":"0.6.1 (a282c32 2021-03-08 build #125)"}
"""

import json
import logging
import os
from pathlib import Path
import shutil
import stat
import sys
import tempfile
import threading
from typing import List, Tuple

import docker
from docker.utils.socket import frames_iter


logging.basicConfig(
    # This is the format used by default in yapapi's examples:
    format="[%(asctime)s %(levelname)s %(name)s] %(message)s",
    level=(logging.DEBUG if os.environ.get("DEBUG_GFTP") else logging.INFO),
)

logger = logging.getLogger("goth.gftp")


# Fixed path at which the volume used to exchange files between this script and
# the `gftp` binary running in the container is mounted in the container
CONTAINER_MOUNT_POINT = "/gftp_volume"


def create_gftp_dirs(requestor_container: str) -> Tuple[Path, Path]:
    """Create a directory with `gftp` script, and a working directory for the script.

    Returns a pair `(script_dir, volume_dir)` where
    - `script_dir` is the directory in which the `gftp` wrapper script is created;
    - `volume_dir` is the directory that should be mounted in the docker container
      running the `gftp` binary at `CONTAINER_MOUNT_POINT`; the wrapper script
      will be set up to read files from, and store them in this directory.
    """
    volume_dir = Path(tempfile.mkdtemp())
    (volume_dir / "in").mkdir()
    (volume_dir / "out").mkdir()

    script_dir = tempfile.mkdtemp()
    script_path = Path(script_dir) / "gftp"
    with open(script_path, "w") as script_file:
        script_file.writelines(
            [
                "#!/bin/sh\n",
                f"python -m goth.gftp {requestor_container} {volume_dir} $*\n",
            ]
        )
    stats = os.stat(script_path)
    os.chmod(script_path, stats.st_mode | stat.S_IEXEC)

    return Path(script_dir), Path(volume_dir)


def _mangle_path(path: Path) -> str:
    """Mangle `path` to a string that can be used as a filename.

    The mangling is deterministic and injective.
    """
    return str(path).replace("SLASH", "SSLASHH").replace(os.sep, "_SLASH_")


def run_gftp_server(gftp_container: str, gftp_volume: Path):
    """Run the `gftp server` command in `gftp_container` and communicate with it.

    Forwards requests read from stdin to the remote command and prints responses
    from the remote command to stdout.

    Makes copies of published files in the `gftp_volume` dir so that they can be
    accessed by the remote command (and modifies requests to reflect modified
    file locations).

    The protocol for communicating with remote command over a socket is described
    here: https://docs.docker.com/engine/api/v1.24/#attach-to-a-container, but
    the link is provided only for reference, we use higher-level utility functions
    from Python Docker SDK here.
    """

    # Start the command and create a socket
    api_client = docker.APIClient()
    exec_id = api_client.exec_create(
        gftp_container, "gftp server", stdin=True, tty=False
    )
    socket = api_client.exec_start(exec_id["Id"], socket=True)._sock

    def container_path_to_volume_path(container_path: Path) -> Path:
        container_path = container_path.resolve()
        relative_path = container_path.relative_to(Path(CONTAINER_MOUNT_POINT))
        return gftp_volume / relative_path

    def volume_path_to_container_path(volume_path: Path) -> Path:
        volume_path = volume_path.resolve()
        relative_path = volume_path.relative_to(gftp_volume)
        return Path(CONTAINER_MOUNT_POINT) / relative_path

    def copy_files_to_volume(files: List[str]) -> List[str]:
        """Copy `files` to the shared volume, return a list of their container paths."""

        container_files = []
        for file in files:
            orig_path = Path(file).resolve()
            copy_path = gftp_volume / "in" / _mangle_path(orig_path)
            shutil.copy(str(orig_path), str(copy_path))
            container_path = volume_path_to_container_path(copy_path)
            container_files.append(str(container_path))
        return container_files

    def response_reader():
        """Read remote command responses from the socket."""

        for stream_type, data in frames_iter(socket, tty=False):
            response = data.decode("utf-8")
            if stream_type == 1:
                # stream_type == 1 means it's from the command's stdout
                logger.debug("received: %s", response.encode("utf-8"))
                try:
                    msg = json.loads(response)
                    if "result" in msg and "file" in msg["result"]:
                        container_file = Path(msg["result"]["file"])
                        volume_file = container_path_to_volume_path(container_file)
                        msg["result"]["file"] = str(volume_file)
                        response = json.dumps(msg) + "\n"
                except Exception as e:
                    logger.warning(
                        "Cannot parse gftp response '%s': %s",
                        response.encode("utf-8"),
                        e,
                    )
                    # Return the response as is
                sys.stdout.write(response)
                sys.stdout.flush()
            elif stream_type == 2:
                # it's from the command's stderr
                logger.debug("stderr: %s", response.strip())
            else:
                raise ValueError(
                    f"Unexpected stream type in a frame header: {stream_type}"
                )

    # Reading responses needs to be done in a separate thread, otherwise it'll block
    reader_thread = threading.Thread(target=response_reader, daemon=True)
    reader_thread.start()

    # Read requests from stdin and write them to the socket
    for line in sys.stdin:
        try:
            msg = json.loads(line)
            method = msg["method"]
            params = msg["params"]
            if method == "publish":
                files = params["files"]
                container_files = copy_files_to_volume(files)
                logger.debug(
                    "replaced `files`; original: %s, new: %s", files, container_files
                )
                params["files"] = container_files
                line = json.dumps(msg) + "\n"
            elif method == "receive":
                output_file = Path(params["output_file"])
                container_file = (
                    Path(CONTAINER_MOUNT_POINT) / "out" / _mangle_path(output_file)
                )
                logger.debug(
                    "replaced `output_file`; original: %s, new: %s",
                    output_file,
                    container_file,
                )
                params["output_file"] = str(container_file)
                line = json.dumps(msg) + "\n"
        except Exception:
            # Pass the request as is
            pass

        logger.debug("sending: (%s)", line.encode("utf-8"))
        socket.sendall(line.encode("utf-8"))

    socket.close()


if __name__ == "__main__":

    container_name = sys.argv[1]
    volume_dir = sys.argv[2]
    command = sys.argv[3]

    if command == "server":
        run_gftp_server(container_name, Path(volume_dir))
    else:
        raise ValueError(f"Command not supported: {command}")



"""Coroutine-based implementation of temporal assertions."""

from goth.assertions.assertions import (
    Assertion,
    AssertionFunction,
    E,
    EventStream,
)

__all__ = [
    "Assertion",
    "AssertionFunction",
    "E",
    "EventStream",
]



"""Interactive runner for `goth` network."""
import asyncio
import logging
from pathlib import Path
from typing import Optional

from goth.configuration import Configuration
from goth.runner import Runner
from goth.runner.probe import ProviderProbe, RequestorProbe


logger = logging.getLogger(__name__)


async def start_network(
    configuration: Configuration,
    log_dir: Optional[Path] = None,
):
    """Start a test network described by `configuration`."""

    runner = Runner(
        base_log_dir=log_dir,
        compose_config=configuration.compose_config,
        test_name="interactive",
        api_assertions_module=None,
        web_root_path=configuration.web_root,
        cancellation_callback=lambda: logger.info("The runner was cancelled"),
    )

    async with runner(configuration.containers):

        providers = runner.get_probes(probe_type=ProviderProbe)
        requestor = runner.get_probes(probe_type=RequestorProbe)[0]

        # Some test steps may be included in the interactive test as well
        for provider in providers:
            await provider.wait_for_offer_subscribed(timeout=10)

        print("\n\033[33;1mNow run your requestor agent as follows:\n")
        env = {"PATH": "$PATH"}
        requestor.set_agent_env_vars(env)
        env_vars = " ".join([f"{key}={val}" for key, val in env.items()])
        subnet = providers[0].provider_agent.subnet
        print(f"$ {env_vars} examples/blender/blender.py --subnet {subnet}")

        print("\nPress Ctrl+C at any moment to stop the test harness.\033[0m\n")

        while True:
            await asyncio.sleep(5)



"""Operators for building temporal assertions."""

import asyncio
from typing import Callable, Optional, TypeVar, TYPE_CHECKING

from goth.assertions import EventStream


if TYPE_CHECKING:

    from typing_extensions import Protocol

    class HasTimestamp(Protocol):
        """A protocol for objects with `timestamp` property."""

        @property
        def timestamp(self) -> float:
            """Return time at which this event occurred."""

    E = TypeVar("E", bound=HasTimestamp)

else:

    E = TypeVar("E")


EventPredicate = Callable[[E], bool]


async def eventually(
    stream: EventStream[E], predicate: EventPredicate, timeout: Optional[float] = None
) -> Optional[E]:
    """Wait for an event that satisfies `predicate` with `timeout`.

    Returns the first event satisfying `predicate` if any such event occurs
    before `timeout`, `None` if the end of events occurs before `timeout` and
    raises `asyncio.TimeoutError` otherwise.
    """
    # This operator could be generalised by:
    # 1) accepting an optional second predicate that would have to be true for each
    #    event *until* `predicate` holds;
    # 2) adding a flag that causes the whole `eventually()` assertion to fail
    #    if the end of events occurs before timeout.

    async def _coro():
        async for e in stream:
            if predicate(e):
                return e
        return None

    return await asyncio.wait_for(_coro(), timeout)



"""Implementation of `yagna id` subcommands."""

from dataclasses import dataclass
from typing import Dict, Optional, Sequence

from goth.runner.cli.base import make_args, parse_json_table, unwrap_ok_err_json
from goth.runner.cli.typing import CommandRunner


@dataclass(frozen=True)
class Identity:
    """Stores information about an identity."""

    alias: Optional[str]
    is_default: bool
    is_locked: bool
    address: str


class YagnaIdMixin:
    """A mixin class that adds support for `<yagna-cmd> id` commands."""

    def id_create(
        self: CommandRunner, data_dir: str = "", alias: str = "", key_file: str = ""
    ) -> Identity:
        """Run `<yagna-cmd> id create` command."""

        args = make_args(
            "id",
            "create",
            "--no-password",
            alias,
            **{"data_dir": data_dir, "from-keystore": key_file},
        )
        output = self.run_json_command(Dict, *args)
        result = unwrap_ok_err_json(output)
        return Identity(
            result["alias"], result["isDefault"], result["isLocked"], result["nodeId"]
        )

    def id_show(
        self: CommandRunner, data_dir: str = "", alias_or_addr: str = ""
    ) -> Optional[Identity]:
        """Return the output of `<yagna-cmd> id show`."""

        args = make_args("id", "show", alias_or_addr, data_dir=data_dir)
        output = self.run_json_command(Dict, *args)
        result = unwrap_ok_err_json(output)
        if result is not None:
            return Identity(
                result["alias"],
                result["isDefault"],
                result["isLocked"],
                result["nodeId"],
            )
        return None

    def id_list(self: CommandRunner, data_dir: str = "") -> Sequence[Identity]:
        """Return the output of `<yagna-cmd> id list`."""

        args = make_args("id", "list", data_dir=data_dir)
        output = self.run_json_command(Dict, *args)
        return [
            Identity(r["alias"], r["default"] == "X", r["locked"] == "X", r["address"])
            for r in parse_json_table(output)
        ]

    def id_update(
        self: CommandRunner,
        alias_or_addr: str,
        data_dir: str = "",
        set_default: bool = False,
    ) -> Identity:
        """Return the output of `<yagna-cmd> id update`."""

        set_default_str = "--set-default" if set_default else None
        args = make_args(
            "id", "update", alias_or_addr, set_default_str, data_dir=data_dir
        )
        output = self.run_json_command(Dict, *args)
        result = unwrap_ok_err_json(output)
        return Identity(
            result["alias"], result["isDefault"], result["isLocked"], result["nodeId"]
        )



"""Module responsible for parsing the docker-compose.yml used in the tests."""
import contextlib
from dataclasses import dataclass
from datetime import datetime
import logging
import os
from pathlib import Path
from typing import ClassVar, Dict, Optional

from docker import DockerClient
import yaml

from goth.runner.container import DockerContainer
from goth.runner.container.build import (
    build_proxy_image,
    build_yagna_image,
    YagnaBuildEnvironment,
)
from goth.runner.container.utils import get_container_address
from goth.runner.exceptions import ContainerNotFoundError
from goth.runner.log import LogConfig
from goth.runner.log_monitor import LogEventMonitor
from goth.runner.process import run_command

logger = logging.getLogger(__name__)

CONTAINER_READY_TIMEOUT = 60  # in seconds
DEFAULT_COMPOSE_FILE = "docker-compose.yml"


@dataclass
class ComposeConfig:
    """Configuration class for `ComposeNetworkManager` instances."""

    build_env: YagnaBuildEnvironment
    """Build environment for building yagna images."""

    file_path: Path
    """Path to the docker-compose.yml file to be used."""

    log_patterns: Dict[str, str]
    """Mapping between service names and regex patterns to look for in their logs.

    The `ComposeNetworkManager` being configured by this object will use the entries
    from this dict to perform ready checks on services. For each service (key)
    the manager will wait for a log line to match the regex pattern (value).
    """


class ComposeNetworkManager:
    """Class which manages a docker-compose network.

    Given the path to a docker-compose.yml file this class can start and stop the
    network, as well as monitor the logs from containers in that network.
    All docker-compose commands are executed in a local shell and not through the
    `DockerClient` used on initialization.
    """

    config: ComposeConfig
    """Configuration for this manager instance."""

    _docker_client: DockerClient
    """Docker client to be used for high-level Docker API calls."""

    _last_compose_path: ClassVar[Optional[Path]] = None
    """Class attribute storing the last compose file used by any manager instance."""

    _log_monitors: Dict[str, LogEventMonitor]
    """Log monitors for containers running as part of docker-compose."""

    _network_gateway_address: str
    """IP address of the gateway for the docker network."""

    def __init__(
        self,
        docker_client: DockerClient,
        config: ComposeConfig,
    ):
        self.config = config
        self.config.file_path = config.file_path.resolve()
        self._docker_client = docker_client
        self._log_monitors = {}
        self._network_gateway_address = ""

    async def start_network(self, log_dir: Path, force_build: bool = False) -> None:
        """Start the compose network based on this manager's compose file.

        This step may include (re)building the network's docker images.
        """
        # Stop the network in case it's already running (e.g. from a previous test)
        await self.stop_network()

        command = ["docker-compose", "-f", str(self.config.file_path), "up", "-d"]

        await build_yagna_image(self.config.build_env)
        await build_proxy_image(self.config.build_env.docker_dir)

        if (
            force_build
            or self.config.file_path != ComposeNetworkManager._last_compose_path
        ):
            command.append("--build")

        await run_command(command, env={**os.environ})
        ComposeNetworkManager._last_compose_path = self.config.file_path

        self._start_log_monitors(log_dir)
        await self._wait_for_containers()
        self._log_running_containers()

    async def _wait_for_containers(self) -> None:
        logger.info("Waiting for compose containers to be ready")
        for name, pattern in self.config.log_patterns.items():
            monitor = self._log_monitors.get(name)
            if not monitor:
                raise RuntimeError(f"No log monitor found for container: {name}")

            logger.debug(
                "Waiting for container to be ready. name=%s, log_pattern=%s",
                name,
                pattern,
            )
            await monitor.wait_for_entry(pattern, timeout=CONTAINER_READY_TIMEOUT)
        logger.info("Compose network ready")

    async def stop_network(self):
        """Stop the running compose network, removing its containers."""
        for name, monitor in self._log_monitors.items():
            logger.debug("stopping log monitor. name=%s", name)
            await monitor.stop()

        await run_command(
            ["docker-compose", "-f", str(self.config.file_path), "down", "-t", "0"]
        )

    def _get_compose_services(self) -> dict:
        """Return services defined in docker-compose.yml."""
        with self.config.file_path.open() as f:
            return yaml.safe_load(f)["services"]

    @property
    def network_gateway_address(self) -> str:
        """Get the IP address of the gateway for the docker network."""

        if not self._network_gateway_address:
            # TODO: parse the docker compose file to get network name,
            # use the default name as fallback
            network_name = DockerContainer.DEFAULT_NETWORK
            network = self._docker_client.networks.get(network_name)
            network_cfg = network.attrs["IPAM"]["Config"][0]
            if "Gateway" in network_cfg:
                self._network_gateway_address = network_cfg["Gateway"]
            else:
                # Use the "subnet" with the last element replaced by "1",
                # e.g. for subnet "172.19.0.0/16" return "172.19.0.1".
                # Not sure if it's totally correct though...
                subnet = network_cfg["Subnet"]
                segments = subnet.split(".")
                assert len(segments) == 4
                segments[3] = "1"
                self._network_gateway_address = ".".join(segments)

        return self._network_gateway_address

    def _log_running_containers(self):
        for container in self._docker_client.containers.list():
            logger.info(
                "[%-25s] IP address: %-15s image: %s",
                container.name,
                get_container_address(self._docker_client, container.name),
                container.image.tags[0],
            )

    def _start_log_monitors(self, log_dir: Path) -> None:
        for service_name in self._get_compose_services():
            log_config = LogConfig(service_name)
            log_config.base_dir = log_dir
            monitor = LogEventMonitor(service_name, log_config)

            containers = self._docker_client.containers.list(
                filters={"name": service_name}
            )
            if not containers:
                raise ContainerNotFoundError(service_name)
            container = containers[0]

            monitor.start(
                container.logs(
                    follow=True, since=datetime.utcnow(), stream=True, timestamps=True
                )
            )
            self._log_monitors[service_name] = monitor


@contextlib.asynccontextmanager
async def run_compose_network(
    compose_manager: ComposeNetworkManager, log_dir: Path, force_build: bool = False
) -> None:
    """Implement AsyncContextManager for starting/stopping docker compose network."""

    try:
        logger.debug(
            "Starting compose network. log_dir=%s, force_build=%s", log_dir, force_build
        )
        await compose_manager.start_network(log_dir, force_build)
        yield
    finally:
        logger.debug("Stopping compose network")
        await compose_manager.stop_network()



"""Mitmproxy addon that traces API calls.

Verifies that a sequence of calls satisfies given properties.
"""
from __future__ import annotations
import logging
from typing import Dict, Optional

from mitmproxy.http import HTTPFlow, HTTPRequest

from goth.api_monitor.api_events import (
    APIEvent,
    APIRequest,
    APIResponse,
    APIError,
)
from goth.assertions.monitor import EventMonitor


class MonitorAddon:
    """This add-on keeps track of API requests and responses."""

    _monitor: EventMonitor[APIEvent]
    _pending_requests: Dict[HTTPRequest, APIRequest]
    _num_requests: int
    _logger: logging.Logger

    def __init__(self, monitor: Optional[EventMonitor[APIEvent]] = None):
        self._monitor = monitor or EventMonitor()
        if not self._monitor.is_running():
            self._monitor.start()
        self._pending_requests = {}
        self._num_requests = 0
        self._logger = logging.getLogger(__name__)

    def _register_event(self, event: APIEvent) -> None:
        """Log an API event and add it to the monitor."""

        self._logger.debug("%s", event)
        self._monitor.add_event_sync(event)

    def request(self, flow: HTTPFlow) -> None:
        """Register a request."""

        self._num_requests += 1
        request = APIRequest(self._num_requests, flow.request)
        self._pending_requests[flow.request] = request
        self._register_event(request)

    def response(self, flow: HTTPFlow) -> None:
        """Register a response."""

        request = self._pending_requests.get(flow.request)
        if request:
            assert flow.response is not None
            response = APIResponse(request, flow.response)
            del self._pending_requests[flow.request]
            self._register_event(response)
        else:
            self._logger.error("Received response for unregistered request: %s", flow)

    def error(self, flow: HTTPFlow) -> None:
        """Register an error."""

        request = self._pending_requests.get(flow.request)
        if request:
            assert flow.error is not None
            error = APIError(request, flow.error, flow.response)
            del self._pending_requests[flow.request]
            self._register_event(error)
        else:
            self._logger.error("Received error for unregistered request: %s", flow)



"""Module for agent components to be used with `Probe` objects."""
import abc
import logging
from typing import Optional, TYPE_CHECKING

from goth.runner.log import LogConfig
from goth.runner.log_monitor import LogEvent, LogEventMonitor
from goth.runner.probe.component import ProbeComponent

if TYPE_CHECKING:
    from goth.runner.probe import Probe

logger = logging.getLogger(__name__)


class AgentComponent(ProbeComponent, abc.ABC):
    """Probe component responsible for managing an agent binary.

    The agent can be an arbitrary binary executed in the container or on host.
    This class also includes logic for monitoring the agent logs and writing
    them to a file.
    """

    log_monitor: LogEventMonitor
    """Monitor and buffer for agent logs, enables asserting for certain lines to be
    present in the log buffer.
    """

    name: str
    """Name of this agent to be used when creating its log file.
    This should be unique when compared to other agents running as part of a
    single test. A good example includes both the probe's name, as well as the
    type of the agent itself, e.g.: `provider_1_ya-provider`.
    """

    def __init__(self, probe: "Probe", name: str):
        super().__init__(probe)
        self.name = name
        self._init_log_monitor()

    @abc.abstractmethod
    async def start(self, *args, **kwargs):
        """Start the agent binary and initialize the internal log monitor."""

    async def stop(self, *args, **kwargs):
        """Stop the agent binary and its log monitor."""
        if self.log_monitor:
            await self.log_monitor.stop()

    def _init_log_monitor(self):
        probe = self.probe
        log_config = LogConfig(file_name=self.name)

        if probe.container.log_config:
            log_config.base_dir = probe.container.log_config.base_dir

        self.log_monitor = LogEventMonitor(self.name, log_config)

    async def wait_for_log(
        self, pattern: str, timeout: Optional[float] = None
    ) -> LogEvent:
        """Search agent logs for a log line with the message matching `pattern`."""
        entry = await self.log_monitor.wait_for_entry(pattern, timeout)
        return entry


class ProviderAgentComponent(AgentComponent):
    """Probe component which runs `ya-provider` in the probe's container."""

    agent_preset: Optional[str]
    """Name of the preset to be used when placing a market offer."""

    subnet: str
    """Name of the subnet to which the provider agent connects."""

    def __init__(self, probe: "Probe", subnet: str, agent_preset: Optional[str] = None):
        super().__init__(probe, f"{probe.name}_ya-provider")
        self.agent_preset = agent_preset
        self.subnet = subnet

    async def start(self):
        """Start the provider agent and attach to its log stream."""
        await super().start()
        probe = self.probe
        probe._logger.info("Starting ya-provider")

        if self.agent_preset:
            probe.container.exec_run(f"ya-provider preset activate {self.agent_preset}")
        probe.container.exec_run(f"ya-provider config set --subnet {self.subnet}")

        log_stream = probe.container.exec_run(
            f"ya-provider run" f" --app-key {probe.app_key} --node-name {probe.name}",
            stream=True,
        )
        self.log_monitor.start(log_stream.output)



"""Built-in web server for serving test assets such as task packages."""

import asyncio
import contextlib
from itertools import cycle
import logging
from pathlib import Path
from typing import Iterator, Optional

from aiohttp import web, web_runner

from goth.address import WEB_SERVER_PORT_END, WEB_SERVER_PORT_START


logger = logging.getLogger(__name__)


class WebServer:
    """A simple web server implemented with the `aiohttp` library."""

    root_path: Path
    """A directory from which the content is served."""

    server_port: int
    """A port on which the server listens."""

    _port_pool: Iterator[int] = cycle(range(WEB_SERVER_PORT_START, WEB_SERVER_PORT_END))
    """Iterator which cycles indefinitely through the range of assigned ports."""

    _server_task: Optional[asyncio.Task] = None
    """An asyncio task wrapping the `aiohttp` server coroutine.

    Not None iff the server is running."""

    _site: web.TCPSite
    """Site object for the TCP socket of this server."""

    def __init__(self, root_path: Path, server_port: Optional[int] = None):
        self.root_path = root_path
        self.server_port = server_port or next(self._port_pool)

    async def _upload_handler(self, request: web.Request) -> web.Response:
        logger.debug("Handling upload request...")
        upload_path = self.root_path / "upload" / request.match_info["filename"]
        with open(upload_path, "wb") as out:
            async for data in request.content.iter_any():
                out.write(data)
        return web.Response()

    async def start(self, server_address: Optional[str]) -> None:
        """Start serving content."""

        if self._server_task:
            logger.warning("Tried to start a web server that is already running")
            return

        app = web.Application()
        app.router.add_put("/upload/{filename}", self._upload_handler)
        app.router.add_static("/", path=Path(self.root_path), name="root")
        runner = web_runner.AppRunner(app)
        await runner.setup()
        self._site = web.TCPSite(runner, server_address, self.server_port)
        self._server_task = asyncio.create_task(self._site.start())
        logger.info(
            "Web server listening on %s:%s, root dir is %s",
            server_address or "*",
            self.server_port,
            self.root_path,
        )

    async def stop(self) -> None:
        """Stop the server."""

        if not self._server_task:
            logger.warning("Tried to stop a web server that is not running")
            return

        if self._server_task.done() and self._server_task.exception():
            await self._server_task

        await self._site.stop()
        self._server_task.cancel()
        await self._server_task
        self._server_task = None
        logger.info("Stopped the web server")


@contextlib.asynccontextmanager
async def run_web_server(server: WebServer, server_address: Optional[str]) -> None:
    """Implement AsyncContextManager protocol for starting/stopping a web server."""

    try:
        logger.debug("Starting web server. address=%s", server_address)
        await server.start(server_address)
        yield
    finally:
        logger.debug("Stopping web server. address=%s", server_address)
        await server.stop()



"""Module related to handling payment IDs in yagna containers."""

from dataclasses import asdict, dataclass
from enum import Enum, unique
import json
from pathlib import Path
import shutil
from tempfile import gettempdir
from typing import Iterator, List, Optional, Sequence
from uuid import uuid4

from goth.project import TEST_DIR

ENV_ACCOUNT_LIST = "ACCOUNT_LIST"

DEFAULT_KEY_DIR = Path(TEST_DIR, "yagna", "keys")


def get_id_directory() -> Path:
    """Return temporary directory used for storing generated payment IDs."""
    temp_id_dir = Path(gettempdir(), "goth_payment_id")
    temp_id_dir.mkdir(exist_ok=True)
    return temp_id_dir


def clean_up() -> None:
    """Perform post-test cleanup related to payment IDs."""
    shutil.rmtree(get_id_directory(), ignore_errors=True)
    get_id_directory().mkdir(exist_ok=True)


class KeyPoolDepletedError(Exception):
    """Error raised when all pre-funded Ethereum keys have been assigned."""

    def __init__(self):
        super().__init__("No more pre-funded Ethereum keys available.")


@unique
# https://docs.python.org/3/library/enum.html#restricted-enum-subclassing
class PaymentDriver(str, Enum):
    """Enum listing the payment drivers that can be used with yagna."""

    erc20 = "erc20"
    zksync = "zksync"


@dataclass
class Account:
    """Data class representing a single yagna payment account."""

    address: str
    driver: PaymentDriver = PaymentDriver.zksync
    network: str = "rinkeby"
    token: str = "tGLM"
    receive: bool = True
    send: bool = True


@dataclass
class EthKey:
    """Data class representing an Ethereum private key."""

    address: str
    crypto: dict
    id: str
    version: int


class PaymentId:
    """Represents a single payment ID to be used with a yagna node.

    Consists of a list of payment accounts along with a common Ethereum key used by
    those accounts as their address.
    Supports dumping the accounts and key to temporary files to be used for mounting
    within a Docker container. For a given payment ID, its files share a common UUID
    in their names.
    """

    accounts: List[Account]
    accounts_file: Path
    key: EthKey
    key_file: Path

    _uuid: str
    """UUID for this instance of `PaymentId`."""

    def __init__(self, accounts: List[Account], key: EthKey):
        self.accounts = accounts
        self.key = key
        self._uuid = uuid4().hex

        self.accounts_file = self._create_accounts_file()
        self.key_file = self._create_key_file()

    def _create_accounts_file(self) -> Path:
        accounts_path = Path(get_id_directory(), f"accounts_{self._uuid}.json")
        with accounts_path.open(mode="w+") as fd:
            serializable = [asdict(a) for a in self.accounts]
            json.dump(serializable, fd)
        return accounts_path

    def _create_key_file(self) -> Path:
        key_path = Path(get_id_directory(), f"key_{self._uuid}.json")
        with key_path.open(mode="w+") as fd:
            json.dump(asdict(self.key), fd)
        return key_path


class PaymentIdPool:
    """Class used for generating yagna payment IDs based on a pool of Ethereum keys.

    The pool of keys is loaded from key files stored in the repo under a given dir
    (`DEFAULT_KEY_DIR` by default).
    """

    _key_pool: Iterator[EthKey]
    """Iterator yielding pre-funded Ethereum keys loaded from a directory."""

    def __init__(self, key_dir: Optional[Path] = None):
        key_dir = key_dir or DEFAULT_KEY_DIR
        self._key_pool = (self._key_from_file(f) for f in key_dir.iterdir())

    def get_id(
        self,
        drivers: Sequence[PaymentDriver] = (PaymentDriver.erc20, PaymentDriver.zksync),
        receive: bool = True,
        send: bool = True,
    ) -> PaymentId:
        """Generate payment accounts with a common, pre-funded Ethereum key.

        Attempts to obtain a key from the pool and, if available, creates a list of
        payment accounts based on the provided parameters.
        For each payment driver specified, a separate account is generated.
        The parameters `receive` and `send` are shared between the accounts.
        Once the key pool is depleted, attempting to get another account results in
        `KeyPoolDepletedError` being raised.
        """
        key = self._get_key()
        account_list = [
            Account(
                address=f"0x{key.address}",
                driver=driver,
                receive=receive,
                send=send,
            )
            for driver in drivers
        ]
        return PaymentId(account_list, key)

    def _get_key(self) -> EthKey:
        try:
            return next(self._key_pool)
        except StopIteration:
            raise KeyPoolDepletedError()

    def _key_from_file(self, path: Path) -> EthKey:
        with path.open() as fd:
            key_dict = json.load(fd)
            return EthKey(**key_dict)



"""Module containing classes related to the yagna REST API client."""
import dataclasses
import logging
from typing import TypeVar, TYPE_CHECKING

from typing_extensions import Protocol

import ya_activity
import ya_market
import ya_payment

from goth.address import (
    ensure_no_trailing_slash,
    ACTIVITY_API_URL,
    MARKET_API_URL,
    PAYMENT_API_URL,
    YAGNA_REST_PORT,
    YAGNA_REST_URL,
)
from goth.runner.probe.component import ProbeComponent

if TYPE_CHECKING:
    from goth.runner.probe import Probe

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ActivityApiClient:
    """
    Client for the activity API of a Yagna daemon.

    The activity API is divided into two domains: control and state. This division is
    reflected in the inner client objects of this class.
    """

    control: ya_activity.RequestorControlApi
    """Client for the control part of the activity API."""

    state: ya_activity.RequestorStateApi
    """Client for the state part of the activity API."""


class ConfigurationProtocol(Protocol):
    """Protocol representing the `Configuration` field of a given REST API module."""

    access_token: str


ConfTVar = TypeVar("ConfTVar", bound=ConfigurationProtocol)
ClientTVar = TypeVar("ClientTVar", covariant=True)


class ApiModule(Protocol[ConfTVar, ClientTVar]):
    """Representation of a REST API module.

    Used for typing `YagnaApiModule._create_api_client`.
    """

    def Configuration(self, host: str) -> ConfTVar:
        """Config instance for this API module."""
        pass

    def ApiClient(self, conf: ConfTVar) -> ClientTVar:
        """Client instance for this API module."""
        pass


class RestApiComponent(ProbeComponent):
    """Component with clients for yagna REST APIs."""

    activity: ActivityApiClient
    """Activity API client."""

    market: ya_market.RequestorApi
    """Market API client."""

    payment: ya_payment.RequestorApi
    """Payment API client."""

    def __init__(self, probe: "Probe"):
        super().__init__(probe)

        # We reach the daemon through MITM proxy running on localhost using the
        # container's unique port mapping
        host_port = probe.container.ports[YAGNA_REST_PORT]
        proxy_ip = "127.0.0.1"
        base_hostname = YAGNA_REST_URL.substitute(host=proxy_ip, port=host_port)

        self._init_activity_api(base_hostname)
        self._init_payment_api(base_hostname)
        self._init_market_api(base_hostname)

    def _create_api_client(
        self, api_module: ApiModule[ConfTVar, ClientTVar], api_url: str
    ) -> ClientTVar:
        api_url = ensure_no_trailing_slash(str(api_url))
        config: ConfTVar = api_module.Configuration(api_url)
        if not self.probe.app_key:
            raise RuntimeError("No app key found. probe=%s", self.probe.name)
        config.access_token = self.probe.app_key
        return api_module.ApiClient(config)

    def _init_activity_api(self, api_base_host: str) -> None:
        api_url = ACTIVITY_API_URL.substitute(base=api_base_host)
        client = self._create_api_client(ya_activity, api_url)
        control = ya_activity.RequestorControlApi(client)
        state = ya_activity.RequestorStateApi(client)
        self.activity = ActivityApiClient(control, state)
        logger.debug("activity API initialized. url=%s", api_url)

    def _init_market_api(self, api_base_host: str) -> None:
        api_url = MARKET_API_URL.substitute(base=api_base_host)
        client = self._create_api_client(ya_market, api_url)
        self.market = ya_market.RequestorApi(client)
        logger.debug("market API initialized. url=%s", api_url)

    def _init_payment_api(self, api_base_host: str) -> None:
        api_url = PAYMENT_API_URL.substitute(base=api_base_host)
        client = self._create_api_client(ya_payment, api_url)
        self.payment = ya_payment.RequestorApi(client)
        logger.debug("payment API initialized. url=%s", api_url)



"""Classes and utilties to manage docker Containers."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Optional

from docker import DockerClient
from docker.models.containers import Container
from transitions import Machine

from goth.runner.log import LogConfig
from goth.runner.log_monitor import LogEventMonitor


@dataclass
class DockerContainerConfig:
    """Configuration to be used for creating a new docker container."""

    name: str
    """Name to be used for this container, must be unique."""

    volumes: Dict[Path, str] = field(default_factory=dict)
    """Volumes to be mounted in the container.

    Keys in this dictionary are paths on the host machine. The values
    are container paths to be used as mount points.
    """

    log_config: Optional[LogConfig] = None
    """Optional custom logging config to be used for this container."""

    privileged_mode: bool = False
    """If set, docker container will be run in privileged mode."""


class State(Enum):
    """Represents states that a Docker container may be in."""

    created = 0
    running = 1
    restarting = 2
    removing = 3
    paused = 4
    exited = 5
    dead = 6


class DockerContainer:
    """A wrapper around `Container`.

    Includes a state machine (from `transitions`) to keep track of a container's
    lifecycle.
    """

    DEFAULT_NETWORK = "docker_default"

    command: List[str]
    """Arguments passed to this container's `command`."""

    entrypoint: str
    """The binary to be run by this container once started."""

    image: str
    """Name of the image to be used for creating this container."""

    logs: Optional[LogEventMonitor]
    """Log buffer for the logs from this container's `entrypoint`."""

    name: str
    """Name to be assigned to this container."""

    network: str
    """Name of the Docker network to be joined once the container is started."""

    # This section lists the members which will be added at runtime by `transitions`
    stop: Callable
    """Stop a running container.

    Internally, this calls `Container.stop` with any kwargs passed here being forwarded
    to that function.
    """

    start: Callable
    """ Start a container which is either created or stopped.

    Internally, this calls `Container.start` with any kwargs passed here being forwarded
    to that function.
    """

    remove: Callable
    """Remove a container.

    This puts this `DockerContainer` in its final state `State.dead`. Internally, this
    calls `Container.remove` with any kwargs passed here being forwarded to that
    function.
    """

    restart: Callable
    """Restart a container.

    Internally, this calls `Container.restart` with any kwargs passed here being
    forwarded to that function.
    """

    _client: DockerClient
    _container: Container
    _state: State

    def __init__(
        self,
        client: DockerClient,
        command: List[str],
        entrypoint: str,
        image: str,
        name: str,
        log_config: Optional[LogConfig] = None,
        network: str = DEFAULT_NETWORK,
        **kwargs,
    ):
        self._client = client
        self.command = command
        self.entrypoint = entrypoint
        self.image = image
        self.name = name
        self.network = network
        self.log_config = log_config
        self.logs = None
        if self.log_config:
            self.logs = LogEventMonitor(self.name, self.log_config)

        self._container = self._client.containers.create(
            self.image,
            entrypoint=self.entrypoint,
            command=self.command,
            detach=True,
            name=self.name,
            network=self.network,
            **kwargs,
        )

        # Initialise the state machine and define allowed transitions
        self._state = State.created
        self.machine = Machine(
            self,
            states=State,
            transitions=[
                {
                    "trigger": "start",
                    "source": [State.created, State.exited],
                    "dest": State.running,
                    "before": self._start,
                },
                {
                    "trigger": "stop",
                    "source": [State.running, State.paused, State.restarting],
                    "dest": State.exited,
                    "before": self._container.stop,
                },
                {
                    "trigger": "remove",
                    "source": "*",
                    "dest": State.dead,
                    "before": self._container.remove,
                },
                {
                    "trigger": "restart",
                    "source": [State.running, State.paused, State.exited],
                    "dest": State.running,
                    "before": self._restart,
                },
            ],
            initial=State.created,
            model_attribute="_state",  # name of the field under which state is stored
            prepare_event="_update_state",  # function to run before each transition
            auto_transitions=False,  # do not generate transition functions
        )

    @property
    def state(self) -> State:
        """State of this container as reported by the Docker daemon."""
        self._update_state()
        return self._state

    def exec_run(self, *args, **kwargs):
        """Proxy to `Container.exec_run`."""
        return self._container.exec_run(*args, **kwargs)

    def _start(self, **kwargs):
        """Start the container."""
        self._container.start(**kwargs)
        if self.logs:
            self.logs.start(self._container.logs(stream=True, follow=True))

    def _restart(self):
        """Restart the container."""
        self._container.restart()
        if self.logs:
            # using naive datetime object as `since` argument deliberately
            # see: https://github.com/docker/docker-py/issues/2712
            self.logs.update_stream(
                self._container.logs(stream=True, follow=True, since=datetime.utcnow())
            )

    def _update_state(self, *_args, **_kwargs):
        """Update the state machine.

        Data is obtained from the Docker daemon by reloading the inner `Container`
        object.
        """

        self._container.reload()
        self.machine.set_state(State[self._container.status])



"""Module with information on the project itself, e.g. project root directory."""
from pathlib import Path

import goth

PROJECT_ROOT = Path(goth.__file__).parent.parent
DOCKER_DIR = PROJECT_ROOT / "docker"
TEST_DIR = PROJECT_ROOT / "test"



"""Classes representing API calls and utility functions."""

import abc
import json
import re
from typing import Optional, Type

from mitmproxy.flow import Error
from mitmproxy.http import HTTPRequest, HTTPResponse

from goth.api_monitor.router_addon import CALLER_HEADER, CALLEE_HEADER


class APIEvent(abc.ABC):
    """Abstract superclass of API event classes."""

    @property
    @abc.abstractmethod
    def timestamp(self) -> float:
        """Return event's time."""

    @property
    @abc.abstractmethod
    def content(self) -> str:
        """Return the content of this event.

        For a request/response event, it's the request/response body.
        For an error event, it's the error message.
        """


class APIRequest(APIEvent):
    """Represents an API request."""

    number: int
    http_request: HTTPRequest

    def __init__(self, number: int, http_request: HTTPRequest):
        self.number = number
        self.http_request = http_request

    @property
    def timestamp(self) -> float:
        """Start time op the `http_request`."""
        return self.http_request.timestamp_start

    @property
    def method(self) -> str:
        """Return the method of the underlying HTTP request."""

        return self.http_request.method

    @property
    def path(self) -> str:
        """Return the method of the underlying HTTP request."""

        return self.http_request.path

    @property
    def caller(self) -> Optional[str]:
        """Return the caller name."""
        return self.http_request.headers.get(CALLER_HEADER)

    @property
    def callee(self) -> Optional[str]:
        """Return the callee name."""
        return self.http_request.headers.get(CALLEE_HEADER)

    @property
    def content(self) -> str:
        """Return the request body."""
        return self.http_request.content.decode("utf-8")

    @property
    def header_str(self) -> str:
        """Return the string representation of this request without the body."""
        return f"{self.caller} -> {self.callee}: {self.method} {self.path}"

    def __str__(self) -> str:
        return f"[request] {self.header_str}; body: {self.content}"


class APIResponse(APIEvent):
    """Represents a response to an API request."""

    request: APIRequest
    http_response: HTTPResponse

    def __init__(self, request: APIRequest, http_response: HTTPResponse):
        self.request = request
        self.http_response = http_response

    @property
    def timestamp(self) -> float:
        """Start time op the `http_response`."""
        return self.http_response.timestamp_start

    @property
    def status_code(self) -> int:
        """Return the HTTP status code."""

        return self.http_response.status_code

    @property
    def content(self) -> str:
        """Return the response body."""
        return self.http_response.content.decode("utf-8")

    def __str__(self) -> str:
        return (
            f"[response ({self.status_code})] "
            f"{self.request.header_str}; body: {self.content}"
        )


class APIError(APIEvent):
    """Represents an error when making an API request or sending a response."""

    request: APIRequest
    error: Error
    http_response: Optional[HTTPResponse]

    def __init__(
        self,
        request: APIRequest,
        error: Error,
        http_response: Optional[HTTPResponse] = None,
    ):
        self.request = request
        self.error = error
        self.http_reponse = http_response

    @property
    def timestamp(self) -> float:
        """Time of the error."""
        return self.error.timestamp

    @property
    def content(self) -> str:
        """Return self."""
        return self.error.msg

    def __str__(self) -> str:
        return f"[error] {self.request.header_str}: {self.content}"


def _match_event(
    event: APIEvent,
    event_class: Type[APIEvent],
    method: Optional[str] = None,
    path_regex: Optional[str] = None,
) -> bool:

    http_request: HTTPRequest
    if isinstance(event, APIRequest):
        http_request = event.http_request
    elif isinstance(event, (APIResponse, APIError)):
        http_request = event.request.http_request
    else:
        return False

    return (
        isinstance(event, event_class)
        and (method is None or http_request.method == method)
        and (path_regex is None or re.search(path_regex, http_request.path) is not None)
    )


def is_create_agreement_request(event: APIEvent) -> bool:
    """Check if `event` is a request of CreateAgreement operation."""

    return _match_event(event, APIRequest, "POST", "^/market-api/v1/agreements$")


def is_collect_demands_request(event: APIEvent, sub_id: str = "") -> bool:
    """Check if `event` is a request of CollectDemants operation."""

    sub_id_re = sub_id if sub_id else "[^/]+"
    return _match_event(
        event, APIRequest, "GET", f"^/market-api/v1/offers/{sub_id_re}/events"
    )


def is_subscribe_offer_request(event: APIEvent) -> bool:
    """Check if `event` is a request of SubscribeOffer operation."""

    return _match_event(event, APIRequest, "POST", "^/market-api/v1/offers$")


def is_unsubscribe_offer_request(event: APIEvent, sub_id: str = "") -> bool:
    """Check if `event` is a request of UnsubscribeOffer operation."""

    sub_id_re = sub_id if sub_id else "[^/]+"
    return _match_event(
        event, APIRequest, "DELETE", f"^/market-api/v1/offers/{sub_id_re}$"
    )


def is_subscribe_offer_response(event: APIEvent) -> bool:
    """Check if `event` is a response of SubscribeOffer operation."""

    return _match_event(event, APIResponse, "POST", "^/market-api/v1/offers$")


def is_invoice_send_response(event: APIEvent) -> bool:
    """Check if `event` is a response for InvoiceSend operation."""

    return _match_event(
        event, APIResponse, "POST", "^/payment-api/v1/provider/invoices/.*/send$"
    )


def get_response_json(event: APIEvent):
    """If `event` is a response then parse and return the included JSON.

    Otherwise return `None`.
    """

    if isinstance(event, APIResponse):
        return json.loads(event.http_response.text)

    return None


def get_activity_id_from_create_response(event: APIEvent) -> Optional[str]:
    """Look for the CreateActivity event to return the ActivityID.

    If `event` is a response to CreateActivity operation then return the activity ID
    included in the response. Otherwise return `None`.
    """

    if (
        isinstance(event, APIResponse)
        and event.request.method == "POST"
        and event.request.path == "/activity-api/v1/activity"
    ):
        return event.http_response.text.strip()[1:-1]

    return None


def get_activity_id_from_delete_response(event: APIEvent) -> Optional[str]:
    """If `event` is a response of DeleteActivity then return the included activity ID.

    Otherwise return `None`.
    """

    if (
        isinstance(event, APIRequest)
        and event.method == "DELETE"
        and event.path.startswith("/activity-api/v1/activity/")
    ):
        return event.path.rsplit("/", 1)[-1]

    return None



"""Exceptions raised by the runner."""


class CommandError(Exception):
    """Error while running a command on a DockerMachine."""

    def __init__(self, message: str):
        super().__init__(message)


class ContainerNotFoundError(Exception):
    """Exception for when the container is not found."""

    def __init__(self, name: str):
        super().__init__(f"No container with name containing '{name}' was found.")


class KeyAlreadyExistsError(CommandError):
    """Specific duplicate key subclass of the CommandError."""


class StopThreadException(Exception):
    """Exception used to stop a running `StoppableThread`."""


class TestFailure(Exception):
    """Base class for errors raised by the test runner when a test fails."""


class StepTimeoutError(TestFailure):
    """Raised on test step timeout."""

    def __init__(self, step: str, time: float):
        super().__init__(f"Step '{step}' timed out after {time:.1f} s")


class TemporalAssertionError(TestFailure):
    """Raised on temporal assertion failure."""

    def __init__(self, assertion: str):
        super().__init__(f"Temporal assertion '{assertion}' failed")


class TimeoutError(Exception):
    """Exception for when a timeout occurs."""



"""Module containing logic for running commands in subprocesses."""

import asyncio
import logging
import subprocess
from typing import Optional, Sequence

from goth.runner.exceptions import CommandError

logger = logging.getLogger(__name__)

RUN_COMMAND_SLEEP_INTERVAL = 0.1  # seconds
RUN_COMMAND_DEFAULT_TIMEOUT = 900  # seconds


async def run_command(
    args: Sequence[str],
    env: Optional[dict] = None,
    log_level: Optional[int] = logging.DEBUG,
    cmd_logger: Optional[logging.Logger] = None,
    log_prefix: Optional[str] = None,
    timeout: float = RUN_COMMAND_DEFAULT_TIMEOUT,
) -> None:
    """Run a command in a subprocess with timeout and logging.

    Lines from stdout and stderr of the subprocess are captured and emitted via
    a locally configured `logging` module logger.

    :param args: sequence consisting of the program to run along with its arguments
    :param env: dict with environment for the command
    :param log_level: logging level at which command output will be logged
    :param cmd_logger: optional logger instance used to log output from the command;
        if not set the default module logger will be used
    :param log_prefix: prefix for log lines with command output; ignored if `cmd_logger`
        is specified. Default: name of the command
    :param timeout: timeout for the command, in seconds. Default: 15 minutes
    """
    logger.info("Running local command: %s", " ".join(args))

    if cmd_logger:
        log_prefix = ""
    else:
        cmd_logger = logger
        if log_prefix is None:
            log_prefix = f"[{args[0]}] "

    async def _run_command():

        proc = await asyncio.subprocess.create_subprocess_exec(
            *args, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        while not proc.stdout.at_eof():
            line = await proc.stdout.readline()
            cmd_logger.log(log_level, "%s%s", log_prefix, line.decode("utf-8").rstrip())

        return_code = await proc.wait()
        if return_code:
            raise CommandError(
                f"Command exited abnormally. args={args}, return_code={return_code}"
            )

    await asyncio.wait_for(_run_command(), timeout=timeout)



"""Classes for running commands inside docker containers."""

from goth.runner.cli.base import DockerJSONCommandRunner
from goth.runner.cli.yagna_app_key_cmd import YagnaAppKeyMixin
from goth.runner.cli.yagna_id_cmd import YagnaIdMixin
from goth.runner.cli.yagna_payment_cmd import YagnaPaymentMixin
from goth.runner.container import DockerContainer


class YagnaDockerCli(
    DockerJSONCommandRunner, YagnaAppKeyMixin, YagnaIdMixin, YagnaPaymentMixin
):
    """A class for running the `yagna` command inside a docker container."""

    def __init__(self, container: DockerContainer):
        super().__init__(container, "yagna")


class Cli:
    """A class for running multiple commands inside a docker container."""

    yagna: YagnaDockerCli
    """A command-line interface for the `yagna` command."""

    def __init__(self, container: DockerContainer):
        self.yagna = YagnaDockerCli(container)



"""Implementation of `yagna app-key` subcommands."""

from dataclasses import dataclass
from typing import Dict, Sequence

from goth.runner.cli.base import make_args, parse_json_table
from goth.runner.cli.typing import CommandRunner
from goth.runner.exceptions import CommandError, KeyAlreadyExistsError


@dataclass(frozen=True)
class AppKeyInfo:
    """Information about an application key."""

    name: str
    key: str
    address: str
    role: str
    created: str


class YagnaAppKeyMixin:
    """A mixin class that adds support for `<yagna-cmd> app-key` commands."""

    def app_key_create(
        self: CommandRunner,
        name: str,
        role: str = "",
        alias_or_addr: str = "",
        data_dir: str = "",
    ) -> str:
        """Run `<cmd> app-key create <name>` with optional extra args.

        Return the application key parsed from the command's output.
        """

        args = make_args(
            "app-key", "create", name, role=role, id=alias_or_addr, data_dir=data_dir
        )
        try:
            output = self.run_json_command(str, *args)
            return output
        except CommandError as ce:
            if "UNIQUE constraint failed: app_key.name" in str(ce):
                raise KeyAlreadyExistsError(name)
            raise ce

    def app_key_drop(
        self: CommandRunner, name: str, address: str = "", data_dir: str = ""
    ) -> str:
        """Run `<cmd> app-key drop <name>` with optional extra args.

        Return the command's output.
        """

        args = make_args("app-key", "drop", name, id=address, data_dir=data_dir)
        return self.run_command(*args)[0]

    def app_key_list(
        self: CommandRunner, address: str = "", data_dir: str = ""
    ) -> Sequence[AppKeyInfo]:
        """Run `<cmd> app-key list` with optional extra args.

        Return the list of `AppKeyInfo`s parsed from the command's output.
        """

        args = make_args("app-key", "list", id=address, data_dir=data_dir)
        output = self.run_json_command(Dict, *args)
        return [
            AppKeyInfo(
                name=info["name"],
                key=info["key"],
                address=info["id"],
                role=info["role"],
                created=info["created"],
            )
            for info in parse_json_table(output)
        ]



"""Helper class for yagna node environment variables and Volumes."""

from typing import Dict

from goth.address import (
    ACTIVITY_API_URL,
    PAYMENT_API_URL,
    ROUTER_HOST,
    ROUTER_PORT,
    YAGNA_BUS_URL,
    YAGNA_REST_URL,
)

DEFAULT_SUBNET = "goth"


def node_environment(
    rest_api_url_base: str = "", account_list: str = ""
) -> Dict[str, str]:
    """Construct an environment for executing commands in a yagna docker container."""

    daemon_env = {
        "CENTRAL_NET_HOST": f"{ROUTER_HOST}:{ROUTER_PORT}",
        "ERC20_RINKEBY_GETH_ADDR": "http://ethereum:8545",
        "GSB_URL": YAGNA_BUS_URL.substitute(host="0.0.0.0"),
        "REQUIRED_CONFIRMATIONS": "1",
        "RINKEBY_TGLM_CONTRACT_ADDRESS": "0xFDFEF9D10d929cB3905C71400ce6be1990EA0F34",
        "RUST_BACKTRACE": "1",
        "RUST_LOG": "debug,tokio_core=info,tokio_reactor=info,hyper=info",
        "YA_PAYMENT_NETWORK": "rinkeby",
        "YAGNA_API_URL": YAGNA_REST_URL.substitute(host="0.0.0.0"),
        "ZKSYNC_FAUCET_ADDR": "http://zksync:3030/zk/donatex",
        "ZKSYNC_RINKEBY_RPC_ADDRESS": "http://zksync:3030",
        # left for compatibility with yagna prior to commit 800efe13
        "ZKSYNC_RPC_ADDRESS": "http://zksync:3030",
    }
    if account_list:
        daemon_env["ACCOUNT_LIST"] = account_list
    node_env = daemon_env

    if rest_api_url_base:
        agent_env = {
            "YAGNA_ACTIVITY_URL": ACTIVITY_API_URL.substitute(base=rest_api_url_base),
            "YAGNA_PAYMENT_URL": PAYMENT_API_URL.substitute(base=rest_api_url_base),
        }
        node_env.update(agent_env)

    return node_env



"""Implementation of `yagna payment` subcommands."""

from dataclasses import dataclass
from typing import Dict, Optional

from goth.runner.cli.base import make_args
from goth.runner.cli.typing import CommandRunner
from goth.runner.container.payment import PaymentDriver


DEFAULT_PAYMENT_DRIVER = PaymentDriver.zksync


@dataclass(frozen=True)
class Payments:
    """Information about payment amounts."""

    accepted: float
    confirmed: float
    rejected: float
    requested: float


@dataclass(frozen=True)
class PaymentStatus:
    """Information about payment status."""

    amount: float
    incoming: Payments
    outgoing: Payments
    reserved: float

    @staticmethod
    def from_dict(source: dict) -> "PaymentStatus":
        """Parse a dict into an instance of `PaymentStatus`."""
        return PaymentStatus(
            amount=float(source["amount"]),
            incoming=Payments(
                **{key: float(value) for key, value in source["incoming"].items()}
            ),
            outgoing=Payments(
                **{key: float(value) for key, value in source["outgoing"].items()}
            ),
            reserved=float(source["reserved"]),
        )


class YagnaPaymentMixin:
    """A mixin class that adds support for `<yagna-cmd> payment` commands."""

    def payment_fund(
        self: CommandRunner, payment_driver: PaymentDriver = DEFAULT_PAYMENT_DRIVER
    ) -> None:
        """Run `<cmd> payment fund` with optional extra args."""
        args = make_args("payment", "fund", driver=payment_driver.name)
        self.run_command(*args)

    def payment_init(
        self: CommandRunner,
        sender_mode: bool = False,
        receiver_mode: bool = False,
        data_dir: str = "",
        payment_driver: PaymentDriver = DEFAULT_PAYMENT_DRIVER,
        address: Optional[str] = None,
        network: Optional[str] = None,
    ) -> None:
        """Run `<cmd> payment init` with optional extra args.

        Return the command's output.
        """

        args = make_args(
            "payment",
            "init",
            data_dir=data_dir,
            driver=payment_driver.name,
            address=address,
            network=network,
        )
        if sender_mode:
            args.append("--sender")
        if receiver_mode:
            args.append("--receiver")

        self.run_command(*args)[0]

    def payment_status(
        self: CommandRunner,
        data_dir: str = "",
        driver: PaymentDriver = DEFAULT_PAYMENT_DRIVER,
    ) -> PaymentStatus:
        """Run `<cmd> payment status` with optional extra args.

        Parse the command's output as a `PaymentStatus` and return it.
        """

        args = make_args("payment", "status", driver.name, data_dir=data_dir)
        output = self.run_json_command(Dict, *args)
        return PaymentStatus.from_dict(output)



"""Utilities related to Docker containers."""
from pathlib import Path
from typing import Dict

from docker import DockerClient

from goth.runner.container import DockerContainer
from goth.runner.exceptions import ContainerNotFoundError


def get_container_address(
    client: DockerClient,
    container_name: str,
    network_name: str = DockerContainer.DEFAULT_NETWORK,
) -> str:
    """Get the IP address of a container in a given network.

    The name of the container does not have to be exact, it may be a substring.
    In case of more than one container name matching the given string, the first
    container is returned, as listed by the Docker daemon.

    Raises `ContainerNotFoundError` if no matching container is found.
    Raises `KeyError` if the container is not connected to the specified network.
    """

    matching_containers = client.containers.list(filters={"name": container_name})
    if not matching_containers:
        raise ContainerNotFoundError(container_name)

    container = matching_containers[0]
    container_networks = container.attrs["NetworkSettings"]["Networks"]
    return container_networks[network_name]["IPAddress"]


def get_volumes_spec(
    volumes: Dict[Path, str], writable: bool = True
) -> Dict[str, dict]:
    """Generate Docker volume specification based on a list of directory mappings.

    `volumes` argument contains mappings between directories. Keys are paths on host,
    values are mount points in the container.
    When `writable` is `True`, the volumes will be mounted as read/write, granting
    the container with write permissions on the mount point.
    """
    return {
        str(host_path.resolve()): {
            "bind": mount_path,
            "mode": "rw" if writable else "ro",
        }
        for host_path, mount_path in volumes.items()
    }



"""This module defines an event monitor class.

that registers events and checks whether temporal assertions are satisfied.
"""

import asyncio
from collections import OrderedDict
import importlib
import logging
import sys
from typing import Callable, Generic, List, Optional, Sequence, Union

from goth.assertions import Assertion, AssertionFunction, E


class MonitorLoggerAdapter(logging.LoggerAdapter):
    """LoggerAdapter adding monitor name to each log message."""

    EXTRA_MONITOR_NAME = "monitor_name"

    _base_logger: logging.Logger

    def __init__(self, base_logger: logging.Logger, *args):
        super().__init__(base_logger, *args)
        self._base_logger = base_logger

    def process(self, msg, kwargs):
        """Process the log message `msg`."""
        return ("[%s] %s" % (self.extra[self.EXTRA_MONITOR_NAME], msg), kwargs)

    @property
    def handlers(self) -> List[logging.Handler]:
        """Return the handlers of the base logger."""
        return self._base_logger.handlers


LogLevel = int


class EventMonitor(Generic[E]):
    """An event monitor.

    That registers *events* (values of type `E`) and manages *assertions* that specify
    properties of sequences of events. The assertions are evaluated with each new
    registered event.
    """

    assertions: "OrderedDict[Assertion[E], LogLevel]"
    """List of all assertions, active or finished.

    For each assertion we also store the log level to be used
    for logging a message when this assertion succeeds.
    """

    name: Optional[str]
    """The name of this monitor, for use in logging."""

    _event_loop: asyncio.AbstractEventLoop
    """The event loop in which this monitor has been started."""

    _events: List[E]
    """List of events registered so far."""

    _incoming: "asyncio.Queue[Optional[E]]"
    """A queue used to pass the events to the worker task."""

    _last_checked_event: int
    """The index of the last event examined by `wait_for_event()` method.

    Subsequent calls to `wait_for_event` will only look at events that occurred
    after this event.
    """

    _logger: Union[logging.Logger, MonitorLoggerAdapter]
    """A logger instance for this monitor."""

    _worker_task: Optional[asyncio.Task]
    """A worker task that registers events and checks assertions."""

    def __init__(
        self,
        name: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
        on_stop=None,
    ) -> None:
        self.assertions = OrderedDict()
        self.name = name

        self._event_loop = asyncio.get_event_loop()
        self._events = []
        self._incoming = asyncio.Queue()
        self._last_checked_event = -1
        self._logger = logger or logging.getLogger(__name__)
        if self.name:
            self._logger = MonitorLoggerAdapter(
                self._logger, {MonitorLoggerAdapter.EXTRA_MONITOR_NAME: self.name}
            )
        self._stop_callback = on_stop
        self._worker_task = None

    def add_assertion(
        self, assertion_func: AssertionFunction[E], log_level: LogLevel = logging.INFO
    ) -> Assertion:
        """Add an assertion function to this monitor."""

        assertion = Assertion(self._events, assertion_func)
        assertion.start()
        self._logger.debug("Assertion '%s' started", assertion.name)
        self.assertions[assertion] = log_level
        return assertion

    def add_assertions(self, assertion_funcs: List[AssertionFunction[E]]) -> None:
        """Add a list of assertion functions to this monitor."""

        for func in assertion_funcs:
            self.add_assertion(func)

    def load_assertions(self, module_name: str) -> None:
        """Load assertion functions from a module."""

        self._logger.info("Loading assertions from module '%s'", module_name)
        mod = importlib.import_module(module_name)
        assert mod is not None
        self.add_assertions(mod.__dict__["TEMPORAL_ASSERTIONS"])

        # Set up the logger in the imported assertions module to use the handlers
        # used by `self._logger`.
        mod_logger = mod.__dict__.get("logger")
        if mod_logger and isinstance(mod_logger, logging.Logger):
            mod_logger.setLevel(self._logger.getEffectiveLevel())
            mod_logger.propagate = False
            for handler in self._logger.handlers:
                if handler not in mod_logger.handlers:
                    mod_logger.addHandler(handler)

    def start(self) -> None:
        """Start tracing events.

        Starting a monitor is decoupled from its initialisation. This allows the
        user to add assertions to the monitor before starting to register events.
        Such assertions are thus guaranteed not to "miss" any event registered
        by the monitor.
        """

        if self.is_running():
            self._logger.warning("Monitor already started")
            return

        self._worker_task = self._event_loop.create_task(self._run_worker())
        self._logger.debug("Monitor started")

    async def add_event(self, event: E) -> None:
        """Register a new event."""

        # Note: this method is `async` even though it does not perform any `await`.
        # This is to ensure that it's directly callable only from code running in
        # an event loop, which in turn guarantees that tasks waiting for input from
        # `self._incoming` will be notified.

        if not self.is_running():
            raise RuntimeError(f"Monitor {self.name or ''} is not running")

        self._incoming.put_nowait(event)

    def add_event_sync(self, event: E) -> None:
        """Schedule registering a new event.

        This function can be called from a thread different from the one
        that started this monitor.
        """

        if not self.is_running():
            raise RuntimeError(f"Monitor {self.name or ''} is not running")

        self._event_loop.call_soon_threadsafe(self._incoming.put_nowait, event)

    async def stop(self) -> None:
        """Stop tracing events."""

        if not self.is_running():
            self._logger.warning("Monitor already stopped")
            return

        self._logger.debug("Stopping the monitor...")
        # This will eventually terminate the worker task:
        self._incoming.put_nowait(None)

        # Set `self._worker_task` to `None` so that when we'll be
        # waiting for the worker task to terminate, `self.is_running()`
        # will return `False` to prevent adding more events.
        worker = self._worker_task
        self._worker_task = None
        assert worker
        await worker
        self._logger.debug("Monitor stopped")

        if not self.finished:
            # This may happen in case of ill-behaved assertions
            self._logger.error("Monitor stopped before all assertions finished")

        if self._stop_callback:
            self._stop_callback()

    def is_running(self) -> bool:
        """Return `True` iff the monitor is accepting events.

        If the worker thread has been terminated by an exception,
        this method re-raises the exception.
        """

        if not self._worker_task:
            return False

        if self._worker_task.done():
            exc = self._worker_task.exception()
            if exc:
                raise exc
            return False

        return True

    def __del__(self) -> None:

        if self.is_running():
            raise RuntimeError("Monitor is still running")

    async def _run_worker(self) -> None:
        """In a loop, register the incoming events and check the assertions."""

        events_ended = False

        while not events_ended:
            event = await self._incoming.get()
            if event is not None:
                self._events.append(event)
            else:
                # `None` is used to signal the end of events
                events_ended = True

            await self._check_assertions(events_ended)

    async def _check_assertions(self, events_ended: bool) -> None:

        event_descr = (
            f"#{len(self._events)} ({self._events[-1]})"
            if not events_ended
            else "EndOfEvents"
        )

        for a, level in list(self.assertions.items()):

            if a.done:
                continue

            await a.update_events(events_ended=events_ended)

            if a.accepted:
                result = a.result()
                msg = "Assertion '%s' succeeded after event: %s; result: %s"
                self._logger.log(level, msg, a.name, event_descr, result)

            elif a.failed:
                await self._report_failure(a, event_descr)

    async def _report_failure(self, a: Assertion, event_descr: str) -> None:
        try:
            a.result()
        except Exception as exc:
            _exc_type, _exc, tb = sys.exc_info()
            # Drop top 3 frames from the traceback: the current one,
            # the one for `a.result()` and the one for the `func_wrapper`
            # used in `__init__()`, so that only the frames of the assertion
            # functions are left.
            for _ in (1, 2, 3):
                tb = tb.tb_next if tb else tb
            msg = "Assertion '%s' failed after event: %s; cause: %s"
            self._logger.error(
                msg, a.name, event_descr, exc, exc_info=(type(exc), exc, tb)
            )

    @property
    def satisfied(self) -> Sequence[Assertion[E]]:
        """Return the satisfied assertions."""

        return [a for a in self.assertions if a.accepted]

    @property
    def failed(self) -> Sequence[Assertion[E]]:
        """Return the failed assertions."""

        return [a for a in self.assertions if a.failed]

    @property
    def done(self) -> Sequence[Assertion[E]]:
        """Return the completed assertions."""

        return [a for a in self.assertions if a.done]

    @property
    def finished(self) -> bool:
        """Return True iif all assertions are done."""

        return all(a.done for a in self.assertions)

    async def wait_for_event(
        self, predicate: Callable[[E], bool], timeout: Optional[float] = None
    ) -> E:
        """Wait for an event that satisfies given `predicate`.

        The first call to this method will examine all events gathered since
        this monitor was started and then, if needed, will wait for up to `timeout`
        seconds for a matching event.

        Subsequent calls will examine all events gathered since the previous call
        returned and then wait for up to `timeout` seconds.

        When `timeout` elapses, `asyncio.TimeourError` will be raised.
        """

        # First examine log lines already seen
        while self._last_checked_event + 1 < len(self._events):
            self._last_checked_event += 1
            event = self._events[self._last_checked_event]
            if predicate(event):
                return event

        # Otherwise create an assertion that waits for a matching event...
        async def wait_for_match(stream) -> E:
            async for e in stream:
                self._last_checked_event = len(stream.past_events) - 1
                if predicate(e):
                    return e
            raise AssertionError("No matching event occurred")

        assertion = self.add_assertion(wait_for_match, logging.DEBUG)

        # ... and wait until the assertion completes
        return await assertion.wait_for_result(timeout=timeout)



"""Type definitions for type-checking mixin classes with mypy (and no other purpose)."""

from typing import TYPE_CHECKING


if TYPE_CHECKING:

    from typing import Tuple, Type, TypeVar

    from typing_extensions import Protocol
    from docker.models.containers import ExecResult

    V = TypeVar("V")

    class CommandRunner(Protocol):
        """A protocol used for type annotations of self in mixin classes."""

        def run_command(self, *cmd_args: str) -> Tuple[str, str]:
            """Run the command with `cmd_args`, raise exception on error."""

        def run_command_no_throw(self, *cmd_args: str) -> ExecResult:
            """Run the command with `cmd_args`, don't raise exceptions."""

        def run_json_command(self, ty: Type[V], *cmd_args: str) -> V:
            """Run the command with `--json` flag."""


else:
    CommandRunner = object



"""Classes and utilities to use a Monitor for log events."""

import asyncio
from datetime import datetime
from enum import Enum
import logging
import re
import time
from typing import Iterator, Optional, Sequence

from func_timeout.StoppableThread import StoppableThread

from goth.assertions.monitor import E, EventMonitor
from goth.runner.exceptions import StopThreadException
from goth.runner.log import LogConfig

logger = logging.getLogger(__name__)


class LogLevel(Enum):
    """Enum representing the rust log levels."""

    ERROR = 1
    WARN = 2
    INFO = 3
    DEBUG = 4
    TRACE = 5


# Pattern to match log lines from the `yagna` binary
pattern = re.compile(
    r"^\[(?P<datetime>[^ ]+) (?P<level>[^ ]+) (?P<module>[^\]]+)\] (?P<message>.*)"
)


class LogEvent:
    """An event representing a log line, used for asserting messages."""

    def __init__(self, log_message: str):
        self._timestamp = None
        self._level = None
        self._module = None
        self._message = log_message
        match = pattern.match(log_message)
        if match:
            result = match.groupdict()

            try:
                formatted_time = datetime.strptime(
                    result["datetime"], "%Y-%m-%dT%H:%M:%SZ"
                )
            except Exception:
                pass
            else:
                self._timestamp = formatted_time.timestamp()
                self._level = LogLevel[result["level"]]
                self._module = result["module"]
                self._message = result["message"]
        if not self._timestamp:
            self._timestamp = time.time()

    @property
    def timestamp(self) -> float:
        """Time of the log message.

        (or time of receiving the event when _module is None)
        """
        return self._timestamp

    @property
    def level(self) -> Optional[LogLevel]:
        """Level reported on the log message.

        Will be empty for multi line logs.
        """
        return self._level

    @property
    def module(self) -> Optional[str]:
        """Source module of this log message.

        Will be empty for multi line logs.
        """
        return self._module

    @property
    def message(self) -> str:
        """Text of the log message."""
        return self._message

    def __repr__(self):
        return (
            f"<LogEvent time={self.timestamp:0.0f}, level={self.level},"
            f" module={self.module}, message={self.message},  >"
        )


def _create_file_logger(config: LogConfig) -> logging.Logger:
    """Create a new file logger configured using the `LogConfig` object provided.

    The target log file will have a .log extension.
    """

    handler = logging.FileHandler(
        (config.base_dir / config.file_name).with_suffix(".log"),
        encoding="utf-8",
        delay=True,
    )
    handler.setFormatter(config.formatter)
    logger_name = f"{config.base_dir}.{config.file_name}"
    logger_ = logging.getLogger(logger_name)
    logger_.setLevel(config.level)
    logger_.addHandler(handler)
    logger_.propagate = False
    return logger_


class PatternMatchingEventMonitor(EventMonitor[E]):
    """An `EventMonitor` that can wait for events that match regex patterns."""

    def event_str(self, event: E) -> str:
        """Return the string associated with `event` on which to perform matching."""
        return str(event)

    async def wait_for_pattern(
        self, pattern: str, timeout: Optional[float] = None
    ) -> E:
        """Wait for an event with string representation matching `pattern`.

        The semantics for this method is as for
        `EventMonitor.wait_for_event(predicate, timeout)`, with `predicate(e)`
        being true iff `event_str(e)` matches `pattern`, for any event `e`.
        """

        regex = re.compile(pattern)
        event = await self.wait_for_event(
            lambda e: regex.match(self.event_str(e)) is not None, timeout
        )
        return event


class LogEventMonitor(PatternMatchingEventMonitor[LogEvent]):
    """Log buffer supporting logging to a file and waiting for a line pattern match.

    `log_config` parameter holds the configuration of the file logger.
    Consecutive values are interpreted as lines by splitting them on the new line
    character.
    Internally it uses a thread to read the stream and add lines to the buffer.
    """

    _buffer_task: Optional[StoppableThread]
    _file_logger: logging.Logger
    _in_stream: Iterator[bytes]

    def __init__(self, name: str, log_config: Optional[LogConfig] = None):
        super().__init__(name)
        if log_config:
            self._file_logger = _create_file_logger(log_config)
        else:
            self._file_logger = logging.getLogger(name)
        self._buffer_task = None
        self._loop = asyncio.get_event_loop()

    def event_str(self, event: LogEvent) -> str:
        """Return the string associated with `event` on which to perform matching."""
        return event.message

    @property
    def events(self) -> Sequence[LogEvent]:
        """Return the events that occurred so far."""
        return self._events

    def start(self, in_stream: Iterator[bytes]):
        """Start reading the logs."""
        super().start()
        self.update_stream(in_stream)
        logger.debug("Started LogEventMonitor. name=%s", self._file_logger.name)

    async def stop(self) -> None:
        """Stop the monitor."""
        await super().stop()
        if self._buffer_task:
            self._buffer_task.stop(StopThreadException)

    def update_stream(self, in_stream: Iterator[bytes]):
        """Update the stream when restarting a container."""
        if self._buffer_task:
            self._buffer_task.stop(StopThreadException)
        self._in_stream = in_stream
        self._buffer_task = StoppableThread(target=self._buffer_input, daemon=True)
        self._buffer_task.start()

    def _buffer_input(self):
        try:
            for chunk in self._in_stream:
                chunk = chunk.decode()
                for line in chunk.splitlines():
                    self._file_logger.info(line)

                    event = LogEvent(line)
                    self.add_event_sync(event)

        except StopThreadException:
            return

    async def wait_for_entry(
        self, pattern: str, timeout: Optional[float] = None
    ) -> LogEvent:
        """Search log for a log entry with the message matching `pattern`.

        The first call to this method will examine all log entries gathered
        since this monitor was started and then, if needed, will wait for
        up to `timeout` seconds (or indefinitely, if `timeout` is `None`)
        for a matching entry.

        Subsequent calls will examine all log entries gathered since
        the previous call returned and then wait for up to `timeout` seconds.
        """
        event = await self.wait_for_pattern(pattern, timeout)
        logger.debug(
            "Log assertion completed with a match. pattern=%s, match=%s",
            pattern,
            event.message,
        )
        return event



"""Classes for running commands in a docker container and some utility functions."""

import json
import logging
import shlex
from typing import Dict, List, Optional, Tuple, Type, TypeVar, TYPE_CHECKING

from docker.models.containers import ExecResult

from goth.runner.exceptions import CommandError

if TYPE_CHECKING:
    from goth.runner.container import DockerContainer


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class DockerCommandRunner:
    """A wrapper for executing a command in a docker container."""

    def __init__(self, container: "DockerContainer", command: str):
        self.container = container
        self.command = command

    def run_command(self, *cmd_args: str) -> Tuple[str, str]:
        """Run the command with `cmd_args`.

        Return a pair of strings containing the standard output and standard error of
        the command on success, or raise `CommandError` on error.
        """

        result = self.run_command_no_throw(*cmd_args)
        # Command's stdout or stderr may be None
        cmd_stderr = result.output[1].decode() if result.output[1] else ""
        cmd_stdout = result.output[0].decode() if result.output[0] else ""
        if result.exit_code != 0:
            raise CommandError(cmd_stderr)
        return cmd_stdout, cmd_stderr

    def run_command_no_throw(self, *cmd_args: str) -> ExecResult:
        """Run the command with `cmd_args`; return its `ExecResult`."""

        cmd_line = f"{self.command} {' '.join(cmd_args)}"
        logger.debug("[%s] command: '%s'", self.container.name, cmd_line)
        return self.container.exec_run(cmd_line, demux=True)


T = TypeVar("T")


class DockerJSONCommandRunner(DockerCommandRunner):
    """Adds method for running commands with `--json` flag."""

    def run_json_command(self, result_type: Type[T], *cmd_args: str) -> T:
        """Add `--json` flag to command arguments and run the command.

        Parse the command output as JSON and return it.
        """

        if "--json" not in cmd_args:
            cmd_args = *cmd_args, "--json"
        cmd_stdout, _ = self.run_command(*cmd_args)
        obj = json.loads(cmd_stdout)
        if isinstance(obj, result_type):
            return obj
        raise CommandError(
            f"Expected a {result_type.__name__} but command returned: {obj}"
        )


def make_args(obj: str, verb: str, *args: str, **opt_args) -> List[str]:
    """Build a list of positional and keyword arguments for a shell command."""

    cmd_args = [obj, verb]
    cmd_args.extend([shlex.quote(arg) for arg in args if arg])
    for key, value in opt_args.items():
        if value:
            cmd_args.extend([f"--{key}", shlex.quote(str(value))])
    return cmd_args


def parse_json_table(output_dict: dict) -> List[Dict[str, str]]:
    """Parse a table in JSON format as returned by some `yagna` subcommands."""

    headers: Optional[list] = output_dict.get("headers")
    values: Optional[list] = output_dict.get("values")

    if not headers or values is None:
        raise ValueError(json.dumps(output_dict))

    result = []
    row: Tuple[str]
    for row in values:
        row_dict = {}
        for i, key in enumerate(headers):
            row_dict[key] = row[i]
        result.append(row_dict)

    return result


U = TypeVar("U")


def unwrap_ok_err_json(output_dict: Dict[str, U]) -> U:
    """Parse `{ "Ok": <result>, "Err": <error>}` JSON; return `<result>`.

    or raise `CommandError(<error>)`.
    """

    if "Ok" in output_dict:
        return output_dict["Ok"]
    if "Err" in output_dict:
        raise CommandError(json.dumps(output_dict["Err"]))
    raise ValueError(json.dumps(output_dict))



"""A class for starting an embedded instance of mitmproxy."""
import asyncio
import contextlib
import logging
import threading
from typing import AsyncIterator, Mapping, Optional

from mitmproxy import options
import mitmproxy.utils.debug
from mitmproxy.tools import _main, cmdline, dump

from goth.address import MITM_PROXY_PORT
from goth.assertions.monitor import EventMonitor
from goth.api_monitor.api_events import APIEvent
from goth.api_monitor.router_addon import RouterAddon
from goth.api_monitor.monitor_addon import MonitorAddon

# This function in `mitmproxy` will try to register signal handlers
# which will fail since the proxy does not run in the main thread.
# So we monkey-patch it to no-op.
mitmproxy.utils.debug.register_info_dumpers = lambda *args: None


logger = logging.getLogger(__name__)


class Proxy:
    """Proxy using mitmproxy to generate events out of http calls."""

    monitor: EventMonitor[APIEvent]
    _proxy_thread: threading.Thread
    _logger: logging.Logger
    _loop: Optional[asyncio.AbstractEventLoop]
    _node_names: Mapping[str, str]
    _server_ready: threading.Event
    """Mapping of IP addresses to node names"""

    _ports: Mapping[str, dict]
    """Mapping of IP addresses to their port mappings"""

    def __init__(
        self,
        node_names: Mapping[str, str],
        ports: Mapping[str, dict],
        assertions_module: Optional[str] = None,
    ):
        self._node_names = node_names
        self._ports = ports
        self._logger = logging.getLogger(__name__)
        self._loop = None
        self._proxy_thread = threading.Thread(
            target=self._run_mitmproxy, name="ProxyThread", daemon=True
        )
        self._server_ready = threading.Event()

        def _stop_callback():
            """Stop `loop` so `proxy_thread` can terminate."""
            if self._loop and self._loop.is_running():
                self._loop.stop()

        self.monitor = EventMonitor("rest", self._logger, on_stop=_stop_callback)
        if assertions_module:
            self.monitor.load_assertions(assertions_module)

    def start(self):
        """Start the proxy thread."""
        self.monitor.start()
        self._proxy_thread.start()
        self._server_ready.wait()

    async def stop(self):
        """Start the proxy monitor and thread."""
        if not self._loop:
            raise RuntimeError("Event loop is not set")
        await self.monitor.stop()
        self._proxy_thread.join()

    def _run_mitmproxy(self):
        """Ran by `self.proxy_thread`."""

        self._loop = asyncio.new_event_loop()
        # Monkey patch the loop to set its `add_signal_handler` method to no-op.
        # The original method would raise error since the loop will run in a non-main
        # thread and hence cannot have signal handlers installed.
        self._loop.add_signal_handler = lambda *args_: None
        asyncio.set_event_loop(self._loop)

        self._logger.info("Starting embedded mitmproxy...")

        # This class is nested since it needs to refer to the `monitor` attribute
        # of the enclosing instance of `Proxy`.
        class MITMProxyRunner(dump.DumpMaster):
            def __init__(inner_self, opts: options.Options) -> None:
                super().__init__(opts)
                inner_self.addons.add(RouterAddon(self._node_names, self._ports))
                inner_self.addons.add(MonitorAddon(self.monitor))

            def start(inner_self):
                self._server_ready.set()
                self._logger.info("Embedded mitmproxy started")
                super().start()

        args = f"-q --mode reverse:http://127.0.0.1 --listen-port {MITM_PROXY_PORT}"
        _main.run(MITMProxyRunner, cmdline.mitmdump, args.split())
        self._logger.info("Embedded mitmproxy exited")


@contextlib.asynccontextmanager
async def run_proxy(proxy: Proxy) -> AsyncIterator[Proxy]:
    """Implement AsyncContextManager protocol for starting and stopping a Proxy."""

    try:
        logger.debug("Starting mitmproxy")
        proxy.start()
        yield
    finally:
        logger.debug("Stopping mitmproxy")
        await proxy.stop()



"""Log utilities for the runner."""

import contextlib
from dataclasses import dataclass
import logging
import logging.config
from pathlib import Path
import tempfile
import time
from typing import Iterator, Optional, Union

import goth
import goth.api_monitor
from goth.assertions.monitor import EventMonitor


DEFAULT_LOG_DIR = Path(tempfile.gettempdir()) / "goth-tests"
FORMATTER_NONE = logging.Formatter("%(message)s")

logger = logging.getLogger(__name__)


class UTCFormatter(logging.Formatter):
    """Custom `logging` formatter that uses `time.gmtime`."""

    converter = time.gmtime


LOGGING_CONFIG = {
    "version": 1,
    "formatters": {
        "none": {"format": "%(message)s"},
        "simple": {"format": "%(levelname)-8s [%(name)-30s] %(message)s"},
        "date": {
            "()": UTCFormatter,
            "format": "%(asctime)s %(levelname)-8s %(name)-30s %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S%z",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "level": "INFO",
        },
        "runner_file": {
            "class": "logging.FileHandler",
            "formatter": "date",
            "filename": "%(base_log_dir)s/runner.log",
            "encoding": "utf-8",
            "level": "DEBUG",
        },
    },
    "loggers": {
        "goth": {
            "handlers": ["console", "runner_file"],
            "propagate": False,
            "level": "DEBUG",
        },
        "goth.api_monitor": {
            "handlers": [],
            "propagate": False,
            # Setting this to "DEBUG" can help in diagnosing issues with routing
            # in the proxy. Use "INFO" to avoid verbose logging of requests/responses.
            "level": "DEBUG",
        },
        "test.yagna": {
            "handlers": ["console", "runner_file"],
            "propagate": False,
            "level": "DEBUG",
        },
        "aiohttp": {
            "handlers": ["console", "runner_file"],
            "level": "DEBUG",
            "propagate": False,
        },
        "transitions": {"level": "WARNING"},
    },
}


def configure_logging(base_dir: Path, console_log_level: Optional[str] = None) -> None:
    """Configure the `logging` module.

    Updates config with `base_dir` before applying the global configuration.
    """

    # substitute `base_log_dir` in `LOGGING_CONFIG` with the actual dir path
    for _name, handler in LOGGING_CONFIG["handlers"].items():
        if "filename" in handler:
            # format the handler's filename with the base dir
            handler["filename"] %= {"base_log_dir": str(base_dir)}

    if console_log_level:
        LOGGING_CONFIG["handlers"]["console"]["level"] = console_log_level

    logging.config.dictConfig(LOGGING_CONFIG)
    logger.info("started logging. dir=%s", base_dir)


@dataclass
class LogConfig:
    """Configuration used to create file loggers."""

    file_name: Union[str, Path]
    base_dir: Path = DEFAULT_LOG_DIR
    formatter: logging.Formatter = FORMATTER_NONE
    level: int = logging.INFO


@contextlib.contextmanager
def configure_logging_for_test(test_log_dir: Path) -> None:
    """Configure loggers to write to files in `test_log_dir`.

    Implements context manager protocol: on entering the context file handlers
    will be added to certain loggers; on exiting these handlers will be removed.
    """

    goth_logger = logging.getLogger(goth.__name__)
    api_monitor_logger = logging.getLogger(goth.api_monitor.__name__)

    runner_handler = None
    proxy_handler = None

    try:
        formatter = logging.Formatter(
            fmt=LOGGING_CONFIG["formatters"]["date"]["format"],
            datefmt=LOGGING_CONFIG["formatters"]["date"]["datefmt"],
        )

        # TODO: ensure the new files created here do not conflict with probe logs
        runner_handler = logging.FileHandler(str(test_log_dir / "test.log"))
        runner_handler.setLevel(logging.DEBUG)
        runner_handler.setFormatter(formatter)
        goth_logger.addHandler(runner_handler)

        proxy_handler = logging.FileHandler(str(test_log_dir / "proxy.log"))
        proxy_handler.setLevel(logging.DEBUG)
        proxy_handler.setFormatter(formatter)
        api_monitor_logger.addHandler(proxy_handler)

        yield

    finally:
        if runner_handler in goth_logger.handlers:
            goth_logger.handlers.remove(runner_handler)
        if proxy_handler in api_monitor_logger.handlers:
            api_monitor_logger.handlers.remove(proxy_handler)


class MonitorHandler(logging.Handler):
    """A logging handler that passes messages from log records to an event monitor."""

    def __init__(self, monitor: EventMonitor[str]):
        self._monitor = monitor
        super().__init__()

    def handle(self, record: logging.LogRecord) -> None:
        """Add the `record`'s message to the associated event monitor."""
        self._monitor.add_event_sync(record.getMessage())


@contextlib.contextmanager
def monitored_logger(name: str, monitor: EventMonitor[str]) -> Iterator[logging.Logger]:
    """Get logger identified by `name` and attach the given event monitor to it.

    The monitor will receive all messages emitted by the logger as events.
    Upon exiting from this context manager, the monitor will be detached
    from the logger.
    """

    logger_ = logging.getLogger(name)
    handler = MonitorHandler(monitor)
    try:
        logger_.addHandler(handler)
        yield logger_
    finally:
        if handler in logger_.handlers:
            logger_.removeHandler(handler)



"""Defines a class representing `goth` configuration."""
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

import yaml

from goth.address import YAGNA_REST_URL, PROXY_HOST
from goth.runner.container.compose import (
    ComposeConfig,
    DEFAULT_COMPOSE_FILE,
    YagnaBuildEnvironment,
)
from goth.runner.container.payment import PaymentIdPool
from goth.node import node_environment
from goth.runner.probe import Probe, YagnaContainerConfig


class Configuration:
    """Configuration of a `goth` test network."""

    containers: List[YagnaContainerConfig]
    """A list of container configurations for nodes in the test network."""

    web_root: Optional[Path]
    """Path to the root directory of the built-in web server.

    `None` means that the web server will not be started.
    """

    compose_config: ComposeConfig
    """Configuration of the underlying docker compose network.

    Contains also the build environment description for yagna containers.
    """

    def __init__(
        self,
        compose_config: ComposeConfig,
        payment_id_pool: PaymentIdPool,
        web_root: Optional[Path] = None,
    ):
        self.containers: List[YagnaContainerConfig] = []
        self.compose_config = compose_config
        self._id_pool: PaymentIdPool = payment_id_pool
        self.web_root = web_root

    def add_node(
        self,
        type: Type[Probe],
        name: str,
        use_proxy: bool,
        privileged_mode: bool,
        volumes: Optional[Dict[Path, Path]],
        **kwargs,
    ) -> None:
        """Add configuration of a new requestor or provider node."""

        if use_proxy:
            node_env = node_environment(
                rest_api_url_base=YAGNA_REST_URL.substitute(host=PROXY_HOST)
            )
        else:
            node_env = node_environment()

        container_cfg = YagnaContainerConfig(
            name=name,
            probe_type=type,
            environment=node_env,
            privileged_mode=privileged_mode,
            subnet="goth",
            volumes=volumes,
            payment_id=self._id_pool.get_id(),
            **kwargs,
        )

        self.containers.append(container_cfg)


class ConfigurationParseError(Exception):
    """An exception raised when parsing a malformed configuration file."""


class _ConfigurationParser:
    """A class for reading `goth` configuration from a `dict` instance."""

    # A type for documents parsed by configuration parser
    Doc = Union[Dict[str, Any], List[Any]]

    def __init__(self, doc: Doc, config_path: Optional[Path], root_key: str = ""):
        self._doc: _ConfigurationParser.Doc = doc
        self._config_path: Optional[Path] = config_path
        self.key: str = root_key

    def __contains__(self, key: Union[int, str]) -> bool:
        return key in self._doc

    def __getitem__(self, key: Union[int, str]) -> Any:
        child_key = self.key + (f".{key}" if isinstance(key, str) else f"[{key}]")
        try:
            value = self._doc[key]
            return (
                _ConfigurationParser(value, self._config_path, child_key)
                if isinstance(value, (dict, list))
                else value
            )
        except KeyError:
            raise ConfigurationParseError(f"Required key is missing: {child_key}")

    def __iter__(self):

        if isinstance(self._doc, dict):
            return iter(self._doc)

        assert isinstance(self._doc, list)
        for index in range(len(self._doc)):
            yield self[index]

    def ensure_type(self, expected: Type) -> None:
        if not isinstance(self._doc, expected):
            raise ConfigurationParseError(
                f"Expected a {expected.__name__} at {self.key}, "
                f"found a {type(self.doc).__name__}"
            )

    @property
    def doc(self) -> Doc:
        """Return the underlying dictionary or list."""
        return self._doc

    def get(self, key: str, default: Any = None) -> Any:
        """Return the value assigned to `key` if defined, otherwise return `default`."""
        return self[key] if key in self else default

    def resolve_path(self, path_str: str) -> Path:
        """Return `path` resolved relative to the directory of the config file."""
        path = Path(path_str).expanduser()
        return (self._config_path.parent / path).resolve()

    def get_path(self, key: str, required: bool = True) -> Optional[Path]:
        """Return a path assigned to `key`, resolved with `self.resolve_path()`."""
        path_opt = self[key] if required else self.get(key)
        return self.resolve_path(path_opt) if path_opt else None

    def read_volumes_spec(self) -> Dict[Path, str]:
        """Read a specification of volumes from this parser's document."""
        self.ensure_type(list)
        volumes = {}
        for mount_spec in self:
            dest = mount_spec["destination"]
            # TODO: "read-only"/"read-write" are the same; both will mount r/w
            # In future, "read-write" should probably copy the asset pointed to
            # by `source` to a temp file/dir, and mount that copied file/dir
            # in a read/write mode
            if "read-only" in mount_spec:
                source = mount_spec.get_path("read-only")
                volumes[source] = dest
            # TODO: "read-write" should
            elif "read-write" in mount_spec:
                source = mount_spec.get_path("read-write")
                volumes[source] = dest
        return volumes

    def read_compose_config(self) -> ComposeConfig:
        """Read a `ComposeConfig` instance from this parser's document."""
        self.ensure_type(dict)

        docker_dir = self.get_path("docker-dir")
        log_patterns = self.get("compose-log-patterns")
        log_patterns.ensure_type(dict)

        build_env_config = self["build-environment"]
        # `build_env_config` may be `None` if no optional build parameters
        # (binary path, commit hash etc.) are specified in the config file
        if build_env_config:
            build_env_config.ensure_type(dict)
            build_env = build_env_config.read_build_env(docker_dir)
        else:
            build_env = YagnaBuildEnvironment(docker_dir)

        return ComposeConfig(
            build_env, docker_dir / DEFAULT_COMPOSE_FILE, log_patterns.doc
        )

    def read_build_env(self, docker_dir: Path) -> YagnaBuildEnvironment:
        """Read a `YagnaBuildEnvironment` instance from this parser's document."""
        self.ensure_type(dict)

        binary_path = self.get_path("binary-path", required=False)
        branch = self.get("branch")
        commit_hash = self.get("commit-hash")
        deb_path = self.get_path("deb-path", required=False)
        release_tag = self.get("release-tag")
        return YagnaBuildEnvironment(
            docker_dir,
            binary_path=binary_path,
            branch=branch,
            commit_hash=commit_hash,
            deb_path=deb_path,
            release_tag=release_tag,
        )


def load_yaml(yaml_path: Union[Path, str]) -> Configuration:
    """Load a configuration from a YAML file at `yaml_path'."""

    import importlib

    with open(str(yaml_path)) as f:
        dict_ = yaml.load(f, yaml.FullLoader)

        network = _ConfigurationParser(dict_, Path(yaml_path))

        key_dir = network.get_path("key-dir")
        compose_config = network["docker-compose"].read_compose_config()
        web_root = network.get_path("web-root", required=False)

        config = Configuration(compose_config, PaymentIdPool(key_dir=key_dir), web_root)

        node_types = {}
        for node_type in network["node-types"]:
            name = node_type["name"]
            type_name = node_type["class"]
            try:
                mod_name, class_name = type_name.rsplit(".", 1)
                module = importlib.import_module(mod_name)
                class_ = module.__dict__[class_name]
                assert isinstance(class_, type)
                assert issubclass(class_, Probe)
            except (AssertionError, KeyError, ValueError):
                raise ConfigurationParseError(
                    f"The value '{type_name}' of {node_type.key + '.class'} "
                    f"does not refer to a subclass of {Probe.__name__}"
                )
            mounts = node_type.get("mount")
            volumes = mounts.read_volumes_spec() if mounts else {}

            privileged_mode = node_type.get("privileged-mode", False)

            node_types[name] = (class_, volumes, privileged_mode)

        for node in network["nodes"]:
            name = node["name"]
            type_name = node["type"]
            use_proxy = node.get("use-proxy", False)
            class_, volumes, privileged_mode = node_types[type_name]
            config.add_node(
                class_,
                name,
                use_proxy=use_proxy,
                privileged_mode=privileged_mode,
                volumes=volumes,
            )

    return config



"""A minimal runner implementation."""
import asyncio
import functools
import logging
import time
from typing import Optional, TYPE_CHECKING

from goth.runner.exceptions import StepTimeoutError

if TYPE_CHECKING:
    from goth.runner.probe import Probe


logger = logging.getLogger(__name__)


def step(default_timeout: float = 10.0):
    """Wrap a step function to implement timeout and log progress."""

    def decorator(func):
        @functools.wraps(func)
        async def wrapper(self: "Probe", *args, timeout: Optional[float] = None):
            timeout = timeout if timeout is not None else default_timeout
            step_name = f"{self.name}.{func.__name__}(timeout={timeout})"
            start_time = time.time()

            logger.info("Running step '%s'", step_name)
            try:
                result = await asyncio.wait_for(func(self, *args), timeout=timeout)
                self.runner.check_assertion_errors()
                step_time = time.time() - start_time
                logger.debug(
                    "Finished step '%s', result: %s, time: %.1f s",
                    step_name,
                    result,
                    step_time,
                )
            except asyncio.TimeoutError:
                step_time = time.time() - start_time
                logger.error("Step '%s' timed out after %.1f s", step_name, step_time)
                raise StepTimeoutError(step_name, step_time)
            except Exception as exc:
                step_time = time.time() - start_time
                logger.error(
                    "Step '%s' raised %s in %.1f",
                    step_name,
                    exc.__class__.__name__,
                    step_time,
                )
                raise
            return result

        return wrapper

    return decorator



"""Package related to downloading assets necessary for building yagna images."""

from abc import ABC
import logging
import os
from pathlib import Path
import re
import shutil
import tempfile
from typing import Any, Callable, Optional

import requests

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(name)-35s %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

ASSET_CACHE_DIR = Path(tempfile.gettempdir()) / "goth_asset_cache"

BASE_URL = "https://api.github.com/repos"

ENV_API_TOKEN = "GITHUB_API_TOKEN"
ENV_YAGNA_BRANCH = "YAGNA_BRANCH"
ENV_YAGNA_COMMIT = "YAGNA_COMMIT_HASH"

DEFAULT_ARTIFACT = "Yagna Linux"
DEFAULT_BRANCH = "master"
DEFAULT_COMMIT = os.getenv(ENV_YAGNA_COMMIT)
DEFAULT_CONTENT_TYPE = "application/vnd.debian.binary-package"
DEFAULT_OWNER = "golemfactory"
DEFAULT_REPO = "yagna"
DEFAULT_TOKEN = os.getenv(ENV_API_TOKEN)
DEFAULT_WORKFLOW = "CI"


class AssetNotFound(Exception):
    """Exception raised when a requested asset could not be found."""


class GithubDownloader(ABC):
    """Base class for downloading assets using GitHub's REST API."""

    repo_url: str
    """Repo URL to be used as base in API requests."""
    session: requests.Session
    """Session object for making HTTP requests."""

    def __init__(
        self,
        owner: str = DEFAULT_OWNER,
        purge_cache: bool = False,
        repo: str = DEFAULT_REPO,
        token: Optional[str] = DEFAULT_TOKEN,
        verbose: bool = False,
    ):
        if not token:
            raise ValueError("GitHub token was not provided.")

        if verbose:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        if purge_cache:
            shutil.rmtree(ASSET_CACHE_DIR)

        self.repo_url = BASE_URL + f"/{owner}/{repo}"
        self.session = requests.Session()
        self.session.headers["Authorization"] = f"token {token}"

    def _cache_get(self, asset_id: str) -> Optional[Path]:
        asset_path: Path = ASSET_CACHE_DIR / asset_id

        if asset_path.is_dir() and list(asset_path.glob("*")):
            return asset_path

        return None

    def _create_cache_dir(self, asset_id: str) -> Path:
        asset_path: Path = ASSET_CACHE_DIR / asset_id
        asset_path.mkdir(exist_ok=True, parents=True)
        return asset_path

    def _parse_link_header(self, header_value: str) -> dict:
        """Parse URLs and their relative positions from a `Link` header value.

        The value of a `Link` header consists of comma-separated tuples, where each
        tuple has a pagination URL and its `rel` attribute.
        `rel` describes its URL's relation to the request the header originates from.
        The value of the `rel` attribute is one of the following:
        `first`, `prev`, `next`, `last`.
        """
        relation_to_url = {}
        links = [link.strip() for link in header_value.split(",")]

        for link in links:
            result = re.search(r'<(\S+)>; rel="(\S+)"', link)
            if not result:
                raise LookupError

            url = result.group(1)
            relation = result.group(2)
            relation_to_url[relation] = url

        return relation_to_url

    def _search_with_pagination(
        self,
        initial_request: requests.PreparedRequest,
        selector: Callable[[requests.Response], Any],
    ):
        """Search response data with `Link` header pagination support.

        First request is made using `initial_request`. Consecutive requests are made
        based on the `Link` header until the last page is reached
        (i.e. no `next` URL is present). The `selector` function is called for each
        response received. If the result from `selector` is non-null, this function
        exits early returning that result.
        """
        response = self.session.send(initial_request)
        logger.debug("_search_with_pagination. initial_url=%s", response.url)

        while True:
            response.raise_for_status()

            result = selector(response)
            if result:
                logger.debug("_search_with_pagination. result=%s", result)
                return result

            relation_to_url = self._parse_link_header(response.headers["Link"])
            logger.debug("_search_with_pagination. relation_to_url=%s", relation_to_url)
            next_url = relation_to_url.get("next")
            if next_url:
                logger.debug("_search_with_pagination. next_url=%s", next_url)
                response = self.session.get(next_url)
            else:
                return None


class ArtifactDownloader(GithubDownloader):
    """Downloader for GitHub Actions artifacts using GitHub's REST API."""

    def _get_workflow(self, workflow_name: str) -> dict:
        """Query the workflow on GitHub Actions."""
        url = f"{self.repo_url}/actions/workflows"
        logger.debug("Fetching workflows. url=%s", url)
        response = self.session.get(url)
        response.raise_for_status()

        workflows = response.json()["workflows"]
        logger.debug("workflows=%s", workflows)
        workflow = next(filter(lambda w: w["name"] == workflow_name, workflows))
        logger.debug("workflow=%s", workflow)
        return workflow

    def _get_latest_run(
        self, workflow: dict, branch: str, commit: Optional[str] = None
    ) -> dict:
        """Filter out the latest workflow run."""
        workflow_id = workflow["id"]
        url = f"{self.repo_url}/actions/workflows/{workflow_id}/runs"
        params = {"status": "completed"}
        if not commit:
            params["branch"] = branch

        request = self.session.prepare_request(
            requests.Request("GET", url, params=params)
        )
        logger.debug("Fetching workflow runs. url=%s", request.url)

        def _filter_workflows(response: requests.Response) -> Optional[dict]:
            workflow_runs = response.json()["workflow_runs"]
            if commit:
                return next(
                    filter(lambda r: r["head_sha"].startswith(commit), workflow_runs),
                    None,
                )
            else:
                return workflow_runs[0]

        workflow_run = self._search_with_pagination(request, _filter_workflows)
        logger.debug("workflow_run=%s", workflow_run)
        return workflow_run

    def _get_artifact(self, artifact_name: str, workflow_run: dict) -> Optional[dict]:
        artifacts_url = workflow_run["artifacts_url"]
        logger.debug("Fetching artifacts. url=%s", artifacts_url)
        response = self.session.get(artifacts_url)
        response.raise_for_status()

        artifacts = response.json()["artifacts"]
        logger.debug("artifacts=%s", artifacts)
        artifact = next(filter(lambda a: artifact_name in a["name"], artifacts), None)

        return artifact

    def _download_artifact(self, artifact: dict) -> Path:
        """Download an artifact from a specific GitHub Actions workflow run.

        Return path to the extracted artifact.
        """
        artifact_id = str(artifact["id"])
        archive_url = artifact["archive_download_url"]

        logger.info("Downloading artifact. url=%s", archive_url)
        with self.session.get(archive_url) as response:
            response.raise_for_status()

            with tempfile.NamedTemporaryFile() as fd:
                fd.write(response.content)
                logger.debug("Extracting zip archive. path=%s", fd.name)
                cache_dir = self._create_cache_dir(artifact_id)
                shutil.unpack_archive(fd.name, format="zip", extract_dir=str(cache_dir))
                logger.debug("Extracted package. path=%s", cache_dir)
                logger.info("Downloaded artifact. url=%s", archive_url)

        return cache_dir

    def download(
        self,
        artifact_name: str = DEFAULT_ARTIFACT,
        branch: str = DEFAULT_BRANCH,
        commit: Optional[str] = DEFAULT_COMMIT,
        output: Optional[Path] = None,
        workflow_name: str = DEFAULT_WORKFLOW,
    ) -> Path:
        """Download an artifact being the result of a given GitHub Actions workflow.

        After downloading, the artifact is extracted and, if specified, saved under
        the directory or file given as `output`.
        Raise `AssetNotFound` if the requested artifact could not be found.
        Return path containing the downloaded artifact.

        :param artifact_name: name of the artifact which should be downloaded
        :param branch: git branch to use when selecting the workflow run
        :param commit: git commit to use when selecting the workflow run
        :param output: directory to which the artifact should be extracted
        :param workflow_name: name of the workflow to select a run from
        """
        logger.debug(
            "ArtifactDownloader#download. name=%s, branch=%s, commit=%s, workflow=%s",
            artifact_name,
            branch,
            commit,
            workflow_name,
        )
        workflow = self._get_workflow(workflow_name)
        latest_run = self._get_latest_run(workflow, branch, commit)
        artifact = self._get_artifact(artifact_name, latest_run)
        if not artifact:
            raise AssetNotFound(f"Artifact not found. name={artifact_name}")

        logger.debug("Found matching artifact. artifact=%s", artifact)
        artifact_id = str(artifact["id"])
        cache_path = self._cache_get(artifact_id)
        if cache_path:
            logger.info("Using cached artifact. cache_path=%s", cache_path)
        else:
            cache_path = self._download_artifact(artifact)

        if output:
            shutil.copytree(cache_path, output, dirs_exist_ok=True)
            logger.debug("Copied artifact to output path. output=%s", str(output))

        return output or cache_path


class ReleaseDownloader(GithubDownloader):
    """Downloader for GitHub repo releases using GitHub's REST API."""

    repo_name: str
    """Name of the repo to download the release from."""

    def __init__(self, repo: str, *args, **kwargs):
        # mypy error here is a bug: https://github.com/python/mypy/issues/6799
        super().__init__(*args, repo=repo, **kwargs)
        self.repo_name = repo

    def _get_latest_release(
        self, tag_substring: str, content_type: str
    ) -> Optional[dict]:
        """Get the latest version, this includes pre-releases.

        Only the versions with `tag_name` that contains `self.tag_substring`
        as a substring are considered.
        """
        url = f"{self.repo_url}/releases"
        logger.debug("Fetching releases. url=%s", url)
        response = self.session.get(url)
        response.raise_for_status()

        all_releases = response.json()
        logger.debug("releases=%s", all_releases)

        def release_filter(release: dict, tag_substring: str) -> bool:
            has_matching_asset = any(
                asset["content_type"] == content_type for asset in release["assets"]
            )
            has_matching_tag = tag_substring in release["tag_name"]
            return has_matching_asset and has_matching_tag

        matching_releases = (
            rel for rel in all_releases if release_filter(rel, tag_substring)
        )
        return next(matching_releases, None)

    def _get_asset(
        self, release: dict, content_type: str, asset_name: Optional[str] = None
    ) -> Optional[dict]:
        assets = release["assets"]
        logger.debug("assets=%s", assets)

        content_assets = filter(lambda a: a["content_type"] == content_type, assets)
        if content_assets and asset_name:
            return next(filter(lambda a: asset_name in a["name"], content_assets), None)
        return next(content_assets, None)

    def _download_asset(self, asset: dict) -> Path:
        """Download an asset from a specific GitHub release."""
        download_url = asset["browser_download_url"]
        asset_id = str(asset["id"])

        logger.info("Downloading asset. url=%s", download_url)
        with self.session.get(download_url) as response:
            response.raise_for_status()
            cache_file = self._create_cache_dir(asset_id) / asset["name"]
            with cache_file.open(mode="wb") as fd:
                fd.write(response.content)
            logger.info("Downloaded asset. path=%s", str(cache_file))

        return cache_file

    def download(
        self,
        asset_name: str = "",
        content_type: str = DEFAULT_CONTENT_TYPE,
        output: Optional[Path] = None,
        tag_substring: str = "",
    ) -> Path:
        """Download the latest release (or pre-release) from a given GitHub repo.

        Raise `AssetNotFound` if the requested release could not be found or if the
        repo has no releases.
        Return path containing the downloaded release.
        :param asset_name: substring the asset's name must contain
        :param content_type: content-type string for the asset to download
        :param output: file path to where the asset should be saved
        :param tag_substring: substring the release's tag name must contain
        """
        release = self._get_latest_release(tag_substring, content_type)
        if not release:
            raise AssetNotFound(
                f"Could not find release. "
                f"repo={self.repo_name}, tag_substring={tag_substring}"
            )

        asset = self._get_asset(release, content_type, asset_name)
        if not asset:
            raise AssetNotFound(
                f"Could not find asset. "
                f"content_type={content_type}, asset_name={asset_name}"
            )

        logger.debug("Found matching asset. name=%s", asset["name"])
        logger.debug("asset=%s", asset)

        asset_id = str(asset["id"])
        cache_path = self._cache_get(asset_id)
        if cache_path:
            cache_path = cache_path / asset["name"]
            logger.info("Using cached release. cache_path=%s", cache_path)
        else:
            cache_path = self._download_asset(asset)

        if output:
            shutil.copy2(cache_path, output)
            logger.debug("Copied release to output path. output=%s", str(output))

        return output or cache_path



"""Common assertions related to API calls."""
from typing import Set

from goth.api_monitor.api_events import (
    APIEvent,
    APIError,
    APIRequest,
    APIResponse,
)

from goth.assertions import EventStream


APIEvents = EventStream[APIEvent]


async def assert_no_api_errors(stream: APIEvents) -> bool:
    """Assert that no instance of `APIError` event ever occurs."""
    async for e in stream:

        if isinstance(e, APIError):
            raise AssertionError(f"API error occurred: {e}")

    return True


async def assert_every_request_gets_response(stream: APIEvents) -> bool:
    """Assert that every request gets a response.

    Assert that for every `APIRequest` event there will eventually occur a
    corresponding `APIResponse` event.
    """
    requests_in_progress: Set[APIRequest] = set()

    async for e in stream:

        if isinstance(e, APIRequest):
            requests_in_progress.add(e)
        elif isinstance(e, APIResponse):
            assert e.request in requests_in_progress
            requests_in_progress.remove(e.request)

    if requests_in_progress:
        a_request = requests_in_progress.pop()
        raise AssertionError(f"request got no response: {a_request}")

    return True



"""Main entry point to `goth`."""
import argparse
import asyncio
from datetime import datetime, timezone
import logging
from pathlib import Path
import shutil

from goth.configuration import load_yaml
from goth.interactive import start_network
from goth.runner.log import configure_logging, DEFAULT_LOG_DIR


DEFAULT_ASSETS_DIR = Path(__file__).parent / "default-assets"


logger = logging.getLogger(__name__)


def make_logs_dir(base_dir: Path) -> Path:
    """Create a unique subdirectory for this test run."""

    date_str = datetime.now(tz=timezone.utc).strftime("%Y%m%d_%H%M%S%z")
    log_dir = base_dir / f"goth_{date_str}"
    log_dir.mkdir(parents=True)

    return log_dir


def start(args):
    """Start the test network using a configuration read from `args.config_file`."""

    configuration = load_yaml(args.config_file)

    base_log_dir = args.log_dir or DEFAULT_LOG_DIR
    log_dir = make_logs_dir(Path(base_log_dir))
    configure_logging(log_dir, args.log_level)

    loop = asyncio.get_event_loop()
    task = start_network(configuration, log_dir=log_dir)
    loop.run_until_complete(task)


def create_config(args):
    """Create sample asset files in the directory specified by `args.output_dir`.

    Will also create the directory if it does not exist yet.
    If any asset file already exists at the given location, it will be overwritten
    if `args.overwrite` is set, otherwise an exception will be raised.
    """

    output_dir = Path(args.output_dir).resolve()

    input_dir = DEFAULT_ASSETS_DIR

    logger.info("Copying default assets from %s to %s", input_dir, output_dir)
    shutil.copytree(input_dir, output_dir, dirs_exist_ok=args.overwrite)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(prog="goth")
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default=logging.INFO,
        help="Log level for the console log",
    )

    subparsers = parser.add_subparsers(
        help="a command to run; add --help to see command-specific options"
    )

    parser_start = subparsers.add_parser("start", help="start a test network")
    parser_start.add_argument(
        "config_file",
        metavar="CONFIG-FILE",
        help="configuration file for a test network",
    )
    parser_start.add_argument(
        "--log-dir", type=str, help="Base directory for goth logs for this run"
    )
    parser_start.set_defaults(function=start)

    parser_cfg = subparsers.add_parser(
        "create-assets", help="create a default network configuration and assets"
    )
    parser_cfg.add_argument(
        "output_dir",
        metavar="OUTPUT-DIR",
        help="target directory for test network configuration and assets",
    )
    parser_cfg.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing asset files"
    )
    parser_cfg.set_defaults(function=create_config)

    args = parser.parse_args()
    try:
        args.function(args)
    except AttributeError:
        parser.print_help()



"""Coroutine-based implementation of temporal assertions."""

import asyncio

from typing import (
    Any,
    AsyncIterator,
    AsyncIterable,
    Callable,
    Coroutine,
    Optional,
    Sequence,
    TypeVar,
    TYPE_CHECKING,
)


E = TypeVar("E")
"""Type variable for the type of events"""


if TYPE_CHECKING:

    from typing_extensions import Protocol

    class EventStream(Protocol, AsyncIterable[E]):
        """A protocol for streams of events of type `E` used by assertion functions."""

        past_events: Sequence[E]
        """A sequence of past events, the last element is the most recent event."""

        events_ended: bool
        """`True` iff there will be no more events."""


else:

    EventStream = AsyncIterable


AssertionFunction = Callable[[EventStream[E]], Coroutine]


class Assertion(AsyncIterable[E]):
    """A class for executing assertion coroutines.

    An instance of this class wraps a coroutine (called the "assertion coroutine")
    and provides an asynchronous generator of events that this coroutine processes.
    After creating an instance of this class, its client should await
    the `update_events()` method each time a new event is appended to the list
    of events (the list is passed as an argument to `Assertion()`).
    After `update_events()` returns, the state of the assertion is updated
    and the client can query that state using the `done`, `accepted`
    and `failed` properties.
    """

    past_events: Sequence[E]
    """See `EventStream`."""

    events_ended: bool
    """See `EventStream`."""

    name: str
    """Assertion name for logging etc."""

    _func: Optional[AssertionFunction]
    """A coroutine function that is executed for this assertion."""

    _task: Optional[asyncio.Task]
    """A task in which the assertion coroutine runs."""

    _ready: Optional[asyncio.Event]
    """An event object used for synchronising the client and the assertion coroutine."""

    _processed: Optional[asyncio.Event]
    """An event object used for synchronising the client and the assertion coroutine."""

    _generator: Optional[AsyncIterator[E]]
    """An asynchronous generator that provides events to the assertion coroutine."""

    def __init__(self, events: Sequence[E], func: AssertionFunction) -> None:
        self.past_events = events
        self.events_ended = False
        self.name = f"{func.__module__}.{func.__name__}"
        self._func = func
        # Creating asyncio objects is decoupled from object initialisation to
        # allow this object to be created and run in different threads (and thus
        # in different event loops).
        self._task = None
        self._ready = None
        self._processed = None
        self._generator = None

    def start(self) -> asyncio.Task:
        """Create asyncio task that runs this assertion."""

        if self.started:
            raise RuntimeError("Assertion already started")

        async def func_wrapper():
            """Ensure `_notify_update_events` is called after processing each event.

            See also comments in `_create_generator()`.
            """
            try:
                return await self._func(self)
            except asyncio.CancelledError:
                raise AssertionError("Assertion cancelled")
            finally:
                self._notify_update_events()

        assert self._func is not None
        self._task = asyncio.create_task(func_wrapper())
        self._ready = asyncio.Event()
        self._processed = asyncio.Event()
        return self._task

    def __str__(self) -> str:
        status = "accepted" if self.accepted else "failed" if self.failed else "ongoing"
        return f"Assertion '{self.name}' ({status})"

    @property
    def started(self) -> bool:
        """Return `True` iff this assertion has started."""
        return self._task is not None

    @property
    def done(self) -> bool:
        """Return `True` iff this assertion finished execution."""
        return self.accepted or self.failed

    @property
    def accepted(self) -> bool:
        """Return `True` iff this assertion finished execution successfuly."""
        return (
            self._task is not None
            and self._task.done()
            and self._task.exception() is None
        )

    @property
    def failed(self) -> bool:
        """Return `True` iff this assertion finished execution by failing."""
        return (
            self._task is not None
            and self._task.done()
            and self._task.exception() is not None
        )

    def result(self) -> Any:
        """Return the result of this assertion.

        The semantics is similar to that of the `result()` method of `asyncio.Task`
        (https://docs.python.org/3/library/asyncio-task.html#task-object):
        If the assertion is done, the result of the assertion coroutine is returned
        or the exception raised by the coroutine is re-raised.
        If the assertion is not done (in particular, if hasn't been started) then
        this method raises `asyncio.InvalidStateError`.
        """
        if not self._task:
            raise asyncio.InvalidStateError("Assertion not started")

        return self._task.result()

    async def wait_for_result(self, timeout: Optional[float] = None) -> Any:
        """Wait until this assertion's result becomes available and return it.

        Optional `timeout` is in seconds, `None` means to wait indefinitely
        (this is the default).
        """
        if not self._task:
            raise asyncio.InvalidStateError("Assertion not started")

        if timeout is None:
            return await self._task

        try:
            return await asyncio.wait_for(self._task, timeout=timeout)
        finally:
            # This is to retrieve exception from `self._task` so no unretrieved
            # exceptions are reported when the event loop closes.
            _ = self._task.exception()

    async def update_events(self, events_ended: bool = False) -> None:
        """Notify the assertion that a new event has been added."""
        if self.events_ended:
            raise AssertionError("Event stream already ended")
        self.events_ended = events_ended

        if self._ready is None or self._processed is None:
            raise asyncio.InvalidStateError("Assertion not started")

        if self.done:
            return

        # This will allow the assertion coroutine to resume execution
        self._ready.set()
        # Here we wait until the assertion coroutine yields control
        await self._processed.wait()
        self._processed.clear()

    def __aiter__(self) -> AsyncIterator[E]:
        """Return an asynchronous generator of events.

        It will yield events to `async for` loops in assertion coroutines.

        For a given assertion `A`, `A.__iter__()` is guaranteed to return
        the same asynchronous generator every time it's called.
        """
        if self._ready is None or self._processed is None:
            raise asyncio.InvalidStateError("Assertion not started")

        if self._generator is None:
            self._generator = self._create_generator()
        return self._generator

    def _notify_update_events(self) -> None:
        """Notify tasks waiting in `update_events()` that the update is processed."""
        if not self._ready or not self._processed:
            raise asyncio.InvalidStateError("Assertion not started")
        self._ready.clear()
        self._processed.set()

    async def _create_generator(self) -> AsyncIterator[E]:
        """Create an asynchronous generator that will be returned by `__aiter__()`."""
        assert self._ready  # to silence mypy

        while True:
            # Wait for `update_events()` to signal that new event is available
            # or that the events ended.
            await self._ready.wait()
            if self.events_ended:
                return

            assert self.past_events
            yield self.past_events[-1]

            # It's important to notify the task waiting in `update_events()`
            # only after the control returns to this generator. In particular,
            # doing this in a `finally` clause wouldn't be correct, as the task
            # waiting in `update_events()` could be notified of processing an event
            # before the assertion coroutine could for example finish execution
            # and mark the assertion as finished.
            #
            # In case the control does not return (since the events end or
            # the assertion coroutine raises an exception), `_notify_update_events()`
            # must be called after returning from the assertion coroutine.
            self._notify_update_events()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['goth',
 'goth.api_monitor',
 'goth.assertions',
 'goth.runner',
 'goth.runner.cli',
 'goth.runner.container',
 'goth.runner.download',
 'goth.runner.probe']

package_data = \
{'': ['*'],
 'goth': ['default-assets/*',
          'default-assets/docker/*',
          'default-assets/keys/*',
          'default-assets/provider/*',
          'default-assets/web-root/*']}

install_requires = \
['aiohttp==3.7.3',
 'docker-compose>=1.27.4,<2.0.0',
 'docker==4.4.0',
 'func_timeout==4.3.5',
 'mitmproxy>=5.1,<6.0',
 'pyyaml>=5.3.1,<6.0.0',
 'transitions>=0.8.1,<0.9.0',
 'typing_extensions>=3.7.4.2,<4.0.0.0',
 'urllib3>=1.25.0,<1.26.0',
 'ya-aioclient>=0.5,<0.6']

setup_kwargs = {
    'name': 'goth',
    'version': '0.1.0',
    'description': 'Golem Test Harness - integration testing framework',
    'long_description': '# Golem Test Harness\n\n![codestyle](https://github.com/golemfactory/goth/workflows/codestyle/badge.svg?event=push)\n![test](https://github.com/golemfactory/goth/workflows/test/badge.svg?event=push)\n\n`goth` is an integration testing framework intended to aid the development process of [`yagna`](https://github.com/golemfactory/yagna) itself, as well as apps built on top of it.\n\n## Running the tests locally\n\n### Python setup\n\n#### Python 3.8\nThe test runner requires Python 3.8+ to be installed on the system. You can check your currently installed Python version by running:\n```\npython3 --version\n```\n\nIf you don\'t have Python installed, download the appropriate package and follow instructions from the [releases page](https://www.python.org/downloads/).\n\nFor the sake of compatibility with other projects and/or your local Python 3 installation you can install [`pyenv`](https://github.com/pyenv/pyenv) to manage and switch between multiple Python versions. The `pyenv` installer can be found [here](https://github.com/pyenv/pyenv-installer).\nYou can also use your preferred way of managing Python virtual environments to achieve this.\n\n#### Project installation\n`goth` is not (yet) available as a standalone package, therefore you will need to set up its development environment in order to use it.\n\n##### poetry\n`goth` uses [`poetry`](https://python-poetry.org/) to manage its dependencies and provide a runner for common tasks (e.g. running E2E tests).\nIf you don\'t have `poetry` available on your system then follow its [installation instructions](https://python-poetry.org/docs/#installation) before proceeding.\nVerify your installation by running:\n```\npoetry --version\n```\n\n##### project dependencies\nTo install the project\'s dependencies run:\n```\npoetry install\n```\n\n### Docker setup\n\n#### Docker Engine\nThe tests are performed on live Yagna nodes running in an isolated network. Currently, this setup is achieved by running a number of Docker containers locally using Docker Compose.\n\nTo run the test containers you will need to have both Docker and Docker Compose installed on your system. To install the Docker engine on your system follow these [instructions](https://docs.docker.com/engine/install/). To verify your installation you can run the `hello-world` Docker image:\n```\ndocker run hello-world\n```\n\n#### Docker Compose\nDocker Compose is a separate binary which needs to be available on your system in order to run Yagna integration tests. `goth` requires `docker-compose` **version 1.27** or higher. There are two ways you can install it:\n1. Download the appropriate executable from the [releases page](https://github.com/docker/compose/releases) and make sure its present on your system\'s `PATH`.\n2. Use the `docker-compose` installed to your `goth` Python environment (you will need to activate the environment in the shell from which you run your tests).\n\n### Running the test network\n\n#### Getting a GitHub API token\nWhen first starting the test network, `goth` uses the GitHub API to fetch metadata and download artifacts and images. Although all of these assets are public, using the GitHub API still requires basic authentication. Therefore, you need to provide `goth` with a personal access token.\n\nTo generate a new token, go to your account\'s [developer settings](https://github.com/settings/tokens).\nYou will need to grant your new token the `public_repo` scope, as well as the `read:packages` scope. The packages scope is required in order to pull Docker images from GitHub.\n\nOnce your token is generated you need to do two things:\n1. Log in to GitHub\'s Docker registry by calling: `docker login docker.pkg.github.com -u {username}`, replacing `{username}` with your GitHub username and pasting in your access token as the password. You only need to do this once on your development machine.\n2. Export an environment variable named `GITHUB_API_TOKEN` and use the access token as its value. This environment variable will need to be available in the shell from which you run `goth`.\n\n### Running the integration tests\nWith project dependencies installed and environment set up you are now ready to launch integration tests.\n\nAll tests related to `yagna` can be found under `test/yagna` with end-to-end tests located in `test/yagna/e2e`. To run them, issue the below command from the project\'s root directory:\n```\npoetry run poe e2e_test\n```\nThe above command makes use of [`poethepoet`](https://github.com/nat-n/poethepoet), a task runner for `poetry`. To see all configured tasks run `poe` with no arguments:\n```\npoetry run poe\n```\n\nFor more granular control (e.g. running one specific test file) you can also invoke pytest directly to run tests:\n```\npytest -svx test/yagna/e2e/test_e2e_wasm.py\n```\n\nThe test runner of choice for `goth` is `pytest`, therefore each test is defined as a separate Python function in a given `.py` file.\n\nEvery test run consists of the following steps:\n1. `docker-compose` is used to start the so-called "static" containers (e.g. local blockchain, HTTP proxy) and create a common Docker network for all containers participating in the test.\n2. The test runner creates a number of Yagna containers (as defined in the test\'s topology) which are connected to the `docker-compose` network.\n3. For each Yagna container started a so-called "probe" object is created and made available inside the test via the `Runner` object.\n4. The integration test scenario is executed as defined in the function called by `pytest`.\n5. Once the test is finished, all previously started Docker containers (both "static" and "dynamic") are removed.\n\n### Custom test options\n\n#### Assets path\nIt\'s possible to provide a custom assets directory which will be mounted in all Yagna containers used for the test. The assets include files such as the exe script definition (`exe-script.json`) or payment configuration (`accounts.json`).\n\nTo override the default path, use the `--assets-path` parameter, passing in the custom path:\n```\npoetry run poe e2e_test --assets-path test/custom_assets/some_directory\n```\n\n#### Log level\nBy default, the test runner will use `INFO` log level. To override it and enable more verbose logging, use the `--log-cli-level` parameter:\n```\npoetry run poe e2e_test --log-cli-level DEBUG\n```\n\n#### Logs path\nThe destination path for all test logs can be overridden using the option `--logs-path`:\n```\npoetry run poe e2e_test --logs-path your/custom/path\n```\n\n#### Yagna binary path\nBy default, a set of yagna binaries is downloaded from GitHub to be used for a given test session. The option `--yagna-binary-path` allows you to use binaries from the local file system instead. Its value must be a path to either a directory tree containing yagna binaries (e.g. `target` directory from a local `cargo` build) or a `.zip` archive file (e.g. downloaded manually from GitHub Actions):\n```\npoetry run poe e2e_test --yagna-binary-path /path/to/binaries\n```\n\n#### Yagna commit hash\nBy default, `goth` uses a `yagna` binary from the latest GitHub Actions successful build on `master` branch. This option can be used to override that behaviour. The value here needs to be a git commit hash being the head for one of the build workflow runs:\n```\npoetry run poe e2e_test --yagna-commit-hash b0ac62f\n```\n\n#### Yagna .deb path\nPath to a local .deb file or a directory containing a number of such archives. All of these .deb files will be installed in the Docker image used for Yagna nodes in the tests. To specify the path, use the option `--yagna-deb-path`:\n```\npoetry run poe e2e_test --yagna-deb-path path/to/yagna.deb\n```\n\n### Troubleshooting integration test runs\nAll components launched during the integration test run record their logs in a pre-determined location. By default, this location is: `$TEMP_DIR/goth-tests`, where `$TEMP_DIR` is the path of the directory used for temporary files. This path will depend either on the shell environment or the operating system on which the tests are being run (see [`tempfile.gettempdir`](https://docs.python.org/3/library/tempfile.html) for more details). This default location can be overridden using the option `--logs-path` when running tests.\n\nThe logs from a test run are recorded in the following directory structure:\n- `runner.log` - logs recorded by the integration test runner engine.\n- `proxy.log` - logs recorded by the HTTP \'sniffer\' proxy, which monitors network traffic going into the yagna daemons launched for the test.\n-  `test_*` - directory with logs generated by nodes running as part of a single test (as defined in the test\'s topology). This directory is named after the name of the test scenario itself.\n The logs follow a straightforward naming convention, with daemons and their clients logging into dedicated files, eg.:\n   - `provider.log` - console output from `provider` yagna daemon.\n   - `provider_agent.log` - console output of the provider agent, running alongside `provider` daemon.\n',
    'author': 'Golem Factory',
    'author_email': 'contact@golem.network',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/golemfactory/goth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Module containing the base class for probe components."""

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from goth.runner.probe import Probe


class ProbeComponent(abc.ABC):
    """Base class for a probe component.

    This serves as a common interface for all classes which can be part of the `Probe`
    class (as in: composition over inheritance).
    """

    def __init__(self, probe: "Probe"):
        self.probe = probe

    probe: "Probe"
    """Probe instance containing this component."""



"""Mitmproxy addon that routes API calls.

Also, adds caller and callee information to request headers.
"""

import logging
from typing import Mapping

from mitmproxy.http import HTTPFlow
from goth.address import (
    HOST_REST_PORT_END,
    HOST_REST_PORT_START,
    YAGNA_REST_PORT,
)


CALLER_HEADER = "X-Caller"
CALLEE_HEADER = "X-Callee"


class RouterAddon:
    """Add-on for mitmproxy to set request headers and route calls.

    This add-on does the following:
    - parses the port on which the original request has been made by API client,
    - sets "X-Caller" and "X-Calle" request headers, so that they can be used
      by subsequent add-ons,
    - routes the request to the appropriate callee, based on the request port.
    """

    _logger: logging.Logger

    _node_names: Mapping[str, str]
    """Mapping of IP addresses to node names"""

    _ports: Mapping[str, dict]
    """Mapping of IP addresses to their port mappings"""

    def __init__(self, node_names: Mapping[str, str], ports: Mapping[str, dict]):
        self._logger = logging.getLogger(__name__)
        self._node_names = node_names
        self._ports = ports

    # pylint: disable = no-self-use
    def request(self, flow: HTTPFlow) -> None:
        """Route the request and set `X-Caller` and `X-Callee` headers."""

        req = flow.request
        self._logger.debug("incoming request %s, headers: %s", req, req.headers)

        try:
            server_addr = req.headers["X-Server-Addr"]
            server_port = int(req.headers["X-Server-Port"])
            remote_addr = req.headers["X-Remote-Addr"]
            node_name = self._node_names[remote_addr]

            if server_port == YAGNA_REST_PORT:
                # It's a provider agent calling a yagna daemon
                # Since we assume that the provider agent runs on the same container
                # as the provider daemon, we route this request to that container's
                # host-mapped daemon port
                req.host = "127.0.0.1"
                req.port = self._ports[remote_addr][server_port]
                req.headers[CALLER_HEADER] = f"{node_name}:agent"
                req.headers[CALLEE_HEADER] = f"{node_name}:daemon"

            elif HOST_REST_PORT_START <= server_port <= HOST_REST_PORT_END:
                # It's a requestor agent calling a yagna daemon.
                # We use localhost as the address together with the original port,
                # since each daemon has its API port mapped to a port on the host
                # chosen from the specified range.
                req.host = "127.0.0.1"
                req.port = server_port
                req.headers[CALLER_HEADER] = f"{node_name}:agent"
                req.headers[CALLEE_HEADER] = f"{node_name}:daemon"

            else:
                flow.kill()
                raise ValueError(f"Invalid server port: {server_port}")

            self._logger.debug(
                "Request from %s for %s:%d/%s routed to %s at %s:%d",
                # request caller:
                req.headers[CALLER_HEADER],
                # original host, port and path:
                server_addr,
                server_port,
                req.path,
                # request recipient:
                req.headers[CALLEE_HEADER],
                # rewritten host and port:
                req.host,
                req.port,
            )

        except (KeyError, ValueError) as ex:
            self._logger.error("Invalid request: %s, error: %s", req, ex.args[0])
            flow.kill()
            raise
